from bloqade.submission.base import SubmissionBackend
import pytest


def test_submission_base():
    A = SubmissionBackend()

    with pytest.raises(NotImplementedError):
        A.cancel_task("1")

    with pytest.raises(NotImplementedError):
        A.task_results("1")

    with pytest.raises(NotImplementedError):
        A.task_status("1")



from typing import Optional
from bloqade.builder.base import Builder


class QuEraService(Builder):
    @property
    def quera(self):
        """
        - Specify Quera backend
        - Possible Next:

            -> `...quera.aquila`
                :: Aquila QPU

            -> `...quera.mock`
                :: mock backend, meant for testings

            -> `...quera.device`
                :: QuEra QPU, specifiy by config_file

        """
        return QuEraDeviceRoute(self)


class QuEraDeviceRoute(Builder):
    def device(self, config_file: Optional[str] = None, **api_config):
        """
        Specify QuEra's QPU device,

        Args:
            config_file (str): file that speficy the target hardware

        Return:
            QuEraHardwareRoutine

        - Possible Next:

            -> `...device().submit`
                :: submit aync remote job

            -> `...device().run`
                :: submit job and wait until job finished
                and results returned

            -> `...device().__callable__`
                :: submit job and wait until job finished
                and results returned


        """
        return self.parse().quera.device(config_file, **api_config)

    def aquila(self):
        """
        Specify QuEra's Aquila QPU

        Return:
            QuEraHardwareRoutine


        - Possible Next:

            -> `...aquila().submit`
                :: submit aync remote job

            -> `...aquila().run`
                :: submit job and wait until job finished
                and results returned

            -> `...aquila().__callable__`
                :: submit job and wait until job finished
                and results returned


        """
        return self.parse().quera.aquila()

    def cloud_mock(self):
        """
        Specify QuEra's Remote Mock QPU

        Return:
            QuEraHardwareRoutine

        - Possible Next:

            -> `...aquila().submit`
                :: submit aync remote job

            -> `...aquila().run`
                :: submit job and wait until job finished
                and results returned

            -> `...aquila().__callable__`
                :: submit job and wait until job finished
                and results returned



        """
        return self.parse().quera.cloud_mock()

    def mock(self, state_file: str = ".mock_state.txt", submission_error: bool = False):
        """
        Specify mock, testing locally.

        Return:
            QuEraHardwareRoutine

        - Possible Next:

            -> `...aquila().submit`
                :: submit aync remote job

            -> `...aquila().run`
                :: submit job and wait until job finished
                and results returned

            -> `...aquila().__callable__`
                :: submit job and wait until job finished
                and results returned



        """
        return self.parse().quera.mock(
            state_file=state_file, submission_error=submission_error
        )



from bloqade.ir.location import Square
import pytest


def test_braket_unsupport_parallel():
    prog = Square(3)

    prog = prog.rydberg.detuning.uniform.piecewise_constant([0.1], [32])
    prog = prog.parallelize(4)

    with pytest.raises(TypeError):
        prog.braket.local_emulator().run(shots=10)



from bloqade.submission.ir.braket import BraketTaskSpecification
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.submission.ir.task_results import QuEraTaskResults, QuEraTaskStatusCode
from typing import Union
from pydantic import BaseModel, Extra
from bloqade.submission.capabilities import get_capabilities
from bloqade.submission.ir.capabilities import QuEraCapabilities


class ValidationError(Exception):
    pass


class SubmissionBackend(BaseModel):
    class Config:
        extra = Extra.forbid

    def get_capabilities(self) -> QuEraCapabilities:
        return get_capabilities()

    def validate_task(
        self, task_ir: Union[BraketTaskSpecification, QuEraTaskSpecification]
    ) -> None:
        raise NotImplementedError

    def submit_task(
        self, task_ir: Union[BraketTaskSpecification, QuEraTaskSpecification]
    ) -> str:
        raise NotImplementedError

    def cancel_task(self, task_id: str) -> None:
        raise NotImplementedError

    def task_results(self, task_id: str) -> QuEraTaskResults:
        raise NotImplementedError

    def task_status(self, task_id: str) -> QuEraTaskStatusCode:
        raise NotImplementedError



from bloqade.ir import (
    rydberg,
    detuning,
    hyperfine,
    Sequence,
    Field,
    Pulse,
    Uniform,
    Linear,
    # ScaledLocations,
    LevelCoupling,
)
from bloqade.ir.control.sequence import NamedSequence
from bloqade.ir import Interval
import pytest
from bloqade import cast
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP
import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_lvlcouple_base():
    lc = LevelCoupling()
    assert lc.children() == []


def test_lvlcouple_hf():
    lc = hyperfine

    assert lc.print_node() == "HyperfineLevelCoupling"

    mystdout = StringIO()
    p = PP(mystdout)
    lc._repr_pretty_(p, 2)

    assert mystdout.getvalue() == "HyperfineLevelCoupling\n"


def test_lvlcouple_ryd():
    lc = rydberg

    assert lc.print_node() == "RydbergLevelCoupling"

    mystdout = StringIO()
    p = PP(mystdout)
    lc._repr_pretty_(p, 2)

    assert mystdout.getvalue() == "RydbergLevelCoupling\n"


def test_sequence():
    # seq empty
    seq = Sequence()
    assert seq.pulses == {}

    # seq non-lvlcoupling:
    with pytest.raises(TypeError):
        Sequence({"c": {}})

    # seq non-lvlcoupling:
    with pytest.raises(TypeError):
        Sequence({rydberg: 3.3})

    # seq full
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    assert seq_full.children() == {"RydbergLevelCoupling": ps}
    assert seq_full.print_node() == "Sequence"

    mystdout = StringIO()
    p = PP(mystdout)
    seq_full._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue() == "Sequence\n"
        "└─ RydbergLevelCoupling\n"
        "   ⇒ Pulse\n"
        "     └─ Detuning\n"
        "        ⇒ Field\n"
        "          └─ Drive\n"
        "             ├─ modulation\n"
        "             │  ⇒ UniformModulation\n"
        "             └─ waveform\n"
        "                ⇒ Linear\n"
        "                  ├─ start\n"
        "                  │  ⇒ Literal: 1.0\n"
        "                  ├─ stop\n"
        "                  │  ⇒ Variable: x\n"
        "                  └─ duration\n"
        "                     ⇒ Literal: 3.0"
    )


def test_named_sequence():
    # seq full
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    # named seq:
    named = NamedSequence(seq_full, "qq")

    assert named.children() == {"sequence": seq_full, "name": "qq"}
    assert named.print_node() == "NamedSequence"

    mystdout = StringIO()
    p = PP(mystdout)
    named._repr_pretty_(p, 10)
    print(repr(mystdout.getvalue()))
    assert (
        mystdout.getvalue() == "NamedSequence\n"
        "├─ sequence\n"
        "│  ⇒ Sequence\n"
        "│    └─ RydbergLevelCoupling\n"
        "│       ⇒ Pulse\n"
        "│         └─ Detuning\n"
        "│            ⇒ Field\n"
        "│              └─ Drive\n"
        "│                 ├─ modulation\n"
        "│                 │  ⇒ UniformModulation\n"
        "│                 └─ waveform\n"
        "│                    ⇒ Linear\n"
        "│                      ├─ start\n"
        "│                      │  ⇒ Literal: 1.0\n"
        "│                      ├─ stop\n"
        "│                      │  ⇒ Variable: x\n"
        "│                      └─ duration\n"
        "│                         ⇒ Literal: 3.0\n"
        "└─ name\n"
        "   ⇒ qq\n"
    )


def test_slice_sequence():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})
    itvl = Interval(cast(0), cast(1.5))

    # slice:
    slc = seq_full[0:1.5]

    assert slc.children() == {"sequence": seq_full, "interval": itvl}
    assert slc.print_node() == "Slice"

    mystdout = StringIO()
    p = PP(mystdout)
    slc._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue() == "Slice\n"
        "├─ sequence\n"
        "│  ⇒ Sequence\n"
        "│    └─ RydbergLevelCoupling\n"
        "│       ⇒ Pulse\n"
        "│         └─ Detuning\n"
        "│            ⇒ Field\n"
        "⋮\n"
        "└─ interval\n"
        "   ⇒ Interval\n"
        "     ├─ start\n"
        "     │  ⇒ Literal: 0\n"
        "     └─ stop\n"
        "        ⇒ Literal: 1.5"
    )


def test_append_sequence():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    app = seq_full.append(seq_full)

    assert app.children() == [seq_full, seq_full]
    assert app.print_node() == "Append"

    mystdout = StringIO()
    p = PP(mystdout)
    app._repr_pretty_(p, 2)
    print(repr(mystdout.getvalue()))

    assert (
        mystdout.getvalue() == "Append\n"
        "├─ Sequence\n"
        "│  └─ RydbergLevelCoupling\n"
        "│     ⇒ Pulse\n"
        "│       └─ Detuning\n"
        "│          ⇒ Field\n"
        "⋮\n"
        "└─ Sequence\n"
        "   └─ RydbergLevelCoupling\n"
        "      ⇒ Pulse\n"
        "        └─ Detuning\n"
        "           ⇒ Field\n"
        "⋮\n"
    )


"""
seq = Sequence(
    {
        rydberg: {
            detuning: {
                Uniform: Linear(start=1.0, stop="x", duration=3.0),
                ScaledLocations({1: 1.0, 2: 2.0}): Linear(
                    start=1.0, stop="x", duration=3.0
                ),
            },
        }
    }
)

print(seq)
print(seq.name("test"))
print(seq.append(seq))
print(seq[:0.5])
"""



from bloqade import start, var
from bloqade.atom_arrangement import Chain
from bloqade.ir.analysis.scan_variables import (
    ScanVariablesAnalogCircuit,
    ScanVariableResults,
)
import numpy as np

from bloqade.ir.control.waveform import to_waveform


def test_1():
    def detuning_wf(t, delta, omega=15):
        return delta * np.sin(omega * t)

    circuit = (
        Chain(15, "lattice_spacing")
        .rydberg.detuning.scale("mask")
        .fn(detuning_wf, "t")
        .amplitude.uniform.constant(15, "t")
        .record("m")
        .phase.location(1, "a")
        .piecewise_linear([0.1, 0.5], [0, 1, 1])
        .parse_circuit()
    )

    scalar_vars = ["t", "omega", "delta", "m", "lattice_spacing", "a"]
    vector_vars = ["mask"]
    expected_result = ScanVariableResults(
        scalar_vars=scalar_vars, vector_vars=vector_vars
    )
    assert expected_result == ScanVariablesAnalogCircuit().emit(circuit)


def test_2():
    t = var("t")

    t_2 = var("T").max(t)
    t_1 = var("T").min(t)

    delta = var("delta") / (2 * np.pi)
    omega_max = var("omega_max") * 2 * np.pi

    @to_waveform(t_2)
    def detuning(t, u):
        return np.abs(t) * u

    circuit = (
        start.add_position(("x", "y"))
        .add_position(("a", "b"))
        .rydberg.rabi.amplitude.scale("mask")
        .piecewise_linear([0.1, t - 0.2, 0.1], [0, omega_max, omega_max, 0])
        .slice(t_1, t_2)
        .uniform.poly([1, 2, 3, 4], t_1)
        .detuning.uniform.constant(10, t_2)
        .uniform.linear(0, delta, t_1)
        .phase.uniform.apply(-(2 * detuning))
        .parallelize(20.0)
        .parse_circuit()
    )

    scalar_vars = ["t", "x", "y", "delta", "T", "omega_max", "a", "u", "b"]
    vector_vars = ["mask"]
    expected_result = ScanVariableResults(
        scalar_vars=scalar_vars, vector_vars=vector_vars
    )
    assert expected_result == ScanVariablesAnalogCircuit().emit(circuit)



Use_bokeh = True

if Use_bokeh:
    from .display import (
        display_ir,
        display_report,
        display_task_ir,
        display_builder,
        figure_ir,
        builder_figure,
        report_figure,
    )
    from .task_visualize import get_task_ir_figure
    from .atom_arragement_visualize import get_atom_arrangement_figure
    from .ir_visualize import get_ir_figure, get_field_figure, get_pulse_figure
else:
    # display
    def display_ir(obj, assignemnts):
        raise Warning("Bokeh not installed", UserWarning)

    def display_report(report):
        raise Warning("Bokeh not installed", UserWarning)

    def display_task_ir(task_ir):
        raise Warning("Bokeh not installed", UserWarning)

    def display_builder(builder, batch_id):
        raise Warning("Bokeh not installed", UserWarning)

    # visualization
    def get_task_ir_figure(task_ir, **fig_kwargs):
        raise Warning("Bokeh not installed", UserWarning)

    # atom arrangement
    def get_atom_arrangement_figure(atom_arng_ir, fig_kwargs=None, **assignments):
        raise Warning("Bokeh not installed", UserWarning)

    # ir
    def get_ir_figure(ir, **assignments):
        raise Warning("Bokeh not installed", UserWarning)

    def get_field_figure(ir_field, title, indicator, **assignments):
        raise Warning("Bokeh not installed", UserWarning)

    def get_pulse_figure(ir_pulse, title: str = None, **assginments):
        raise Warning("Bokeh not installed", UserWarning)

    # figure:
    def figure_ir(obj, assignemnts):
        raise Warning("Bokeh not installed", UserWarning)

    def builder_figure(builder, batch_id):
        raise Warning("Bokeh not installed", UserWarning)

    def report_figure(report):
        raise Warning("Bokeh not installed", UserWarning)



from bloqade.builder.base import Builder
from bloqade.builder.sequence_builder import SequenceBuilder
from bloqade.builder.drive import Drive
from bloqade.ir.control.sequence import Sequence
from beartype import beartype


class ProgramStart(Drive, Builder):
    """
    ProgramStart is the base class for a starting/entry node for building a program.
    """

    @beartype
    def apply(self, sequence: Sequence) -> SequenceBuilder:
        """
        Apply a pre-built sequence to a program.

        This allows you to build a program independent of any geometry
        and then `apply` the program to said geometry. Or, if you have a
        program you would like to try on multiple geometries you can
        trivially do so with this.

        Example Usage:
        ```
        >>> from numpy import pi
        >>> seq = start.rydberg.rabi.amplitude.constant(2.0 * pi, 4.5)
        # choose a geometry of interest to apply the program on
        >>> from bloqade.atom_arrangement import Chain, Kagome
        >>> complete_program = Chain(10).apply(seq)
        # you can .apply to as many geometries as you like
        >>> another_complete_program = Kagome(3).apply(seq)
        ```

        - From here you can now do:
            - `...assign(assignments).bloqade`: select the bloqade
                local emulator backend
            - `...assign(assignments).braket`: select braket
                local emulator or QuEra hardware
            - `...assign(assignments).device(specifier_string)`: select
                backend by specifying a string
        - Assign multiple values to a single variable for a parameter sweep:
            - `...assign(assignments).batch_assign(assignments)`:
        - Parallelize the program register, duplicating the geometry and waveform
            sequence to take advantage of all available
          space/qubits on the QPU:
            - `...assign(assignments).parallelize(cluster_spacing)`
        - Defer value assignment of certain variables to runtime:
            - `...assign(assignments).args([previously_defined_vars])`

        """
        return SequenceBuilder(sequence, self)



from pydantic.dataclasses import dataclass
from bloqade.ir.location.base import AtomArrangement, LocationInfo, SiteFilling
from bloqade.builder.typing import ScalarType
from beartype.typing import List, Tuple, Union
from beartype import beartype


@dataclass(init=False)
class ListOfLocations(AtomArrangement):
    __match_args__ = ("location_list",)
    location_list: List[LocationInfo]

    @beartype
    def __init__(
        self,
        location_list: List[Union[LocationInfo, Tuple[ScalarType, ScalarType]]] = [],
    ):
        self.location_list = []
        for ele in location_list:
            if isinstance(ele, LocationInfo):
                self.location_list.append(ele)
            else:
                self.location_list.append(LocationInfo(ele, True))

        if self.location_list:
            self.__n_atoms = sum(
                1 for loc in self.location_list if loc.filling == SiteFilling.filled
            )
            self.__n_sites = len(self.location_list)
            self.__n_vacant = self.__n_sites - self.__n_atoms
            self.__n_dims = len(self.location_list[0].position)
        else:
            self.__n_sites = 0
            self.__n_atoms = 0
            self.__n_dims = None

        super().__init__()

    @property
    def n_atoms(self):
        return self.__n_atoms

    @property
    def n_sites(self):
        return self.__n_sites

    @property
    def n_vacant(self):
        return self.__n_vacant

    @property
    def n_dims(self):
        return self.__n_dims

    def enumerate(self):
        return iter(self.location_list)

    def __iter__(self):
        return iter(self.location_list)


start = ListOfLocations()
"""
A Program starting point, alias of empty
[`ListOfLocations`][bloqade.ir.location.list.ListOfLocations].

- Next possible steps to build your program are:
- Specify which level coupling to address with:
    - `start.rydberg`: for [`Rydberg`][bloqade.builder.coupling.Rydberg]
        Level coupling
    - `start.hyperfine`: for [`Hyperfine`][bloqade.builder.coupling.Hyperfine]
        Level coupling
    - LOCKOUT: You cannot add atoms to your geometry after specifying level coupling.
- continue/start building your geometry with:
    - `start.add_position()`: to add atom(s) to current register. It will accept:
        - A single coordinate, represented as a tuple (e.g. `(5,6)`) with a value that
          can either be:
            - integers: `(5,6)`
            - floats: `(5.1, 2.5)`
            - strings (for later variable assignment): `("x", "y")`
            - [`Scalar`][bloqade.ir.scalar.Scalar] objects: `(2*cast("x"), 5+cast("y"))`
        - A list of coordinates, represented as a list of types mentioned previously.
        - A numpy array with shape (n, 2) where n is the total number of atoms
"""



from bloqade.builder.base import Builder
from bloqade.builder.coupling import LevelCoupling, Rydberg, Hyperfine
from bloqade.builder.sequence_builder import SequenceBuilder
from bloqade.builder.field import Field, Detuning, RabiAmplitude, RabiPhase
from bloqade.builder.spatial import SpatialModulation, Location, Uniform, Scale
from bloqade.builder.waveform import WaveformPrimitive, Slice, Record, Sample, Fn
from bloqade.builder.assign import Assign, BatchAssign, ListAssign
from bloqade.builder.args import Args
from bloqade.builder.parallelize import Parallelize
from bloqade.builder.parse.stream import BuilderNode, BuilderStream
import bloqade.ir as ir
from beartype.typing import TYPE_CHECKING, Tuple, Union, Dict, List, Optional, Set

if TYPE_CHECKING:
    from bloqade.ir.routine.params import ParamType
    from bloqade.ir.routine.base import Routine
    from bloqade.ir.analog_circuit import AnalogCircuit


class Parser:
    stream: Optional["BuilderStream"] = None
    vector_node_names: Set[str] = set()
    sequence: ir.Sequence = ir.Sequence()
    register: Union[ir.AtomArrangement, ir.ParallelRegister, None] = None
    batch_params: List[Dict[str, "ParamType"]] = [{}]
    static_params: Dict[str, "ParamType"] = {}
    order: Tuple[str, ...] = ()

    def reset(self, builder: Builder):
        self.stream = BuilderStream.create(builder)
        self.vector_node_names = set()
        self.sequence = ir.Sequence()
        self.register = None
        self.batch_params = [{}]
        self.static_params = {}
        self.order = ()

    def read_address(self, stream) -> Tuple[LevelCoupling, Field, BuilderNode]:
        spatial = stream.read_next([Location, Uniform, Scale])
        curr = spatial

        if curr is None:
            return (None, None, None)

        while curr.next is not None:
            if not isinstance(curr.node, SpatialModulation):
                break
            curr = curr.next

        if type(spatial.node.__parent__) in [Detuning, RabiAmplitude, RabiPhase]:
            field = spatial.node.__parent__  # field is updated
            if type(field) in [RabiAmplitude, RabiPhase]:
                coupling = field.__parent__.__parent__  # skip Rabi
            else:
                coupling = field.__parent__

            # coupling not updated
            if type(coupling) not in [Rydberg, Hyperfine]:
                coupling = None
            return (coupling, field, spatial)
        else:  # only spatial is updated
            return (None, None, spatial)

    def read_waveform(self, head: BuilderNode) -> Tuple[ir.Waveform, BuilderNode]:
        curr = head
        waveform = None
        while curr is not None:
            node = curr.node

            if isinstance(node, Slice):
                waveform = waveform[node._start : node._stop]
            elif isinstance(node, Record):
                waveform = waveform.record(node._name)
            elif isinstance(node, Sample):
                interpolation = node._interpolation
                if interpolation is None:
                    if self.field_name == ir.rabi.phase:
                        interpolation = ir.Interpolation.Constant
                    else:
                        interpolation = ir.Interpolation.Linear
                fn_waveform = node.__parent__.__bloqade_ir__()
                sample_waveform = ir.Sample(fn_waveform, interpolation, node._dt)
                if waveform is None:
                    waveform = sample_waveform
                else:
                    waveform = waveform.append(sample_waveform)
            elif (
                isinstance(node, Fn)
                and curr.next is not None
                and isinstance(curr.next.node, Sample)
            ):
                pass
            elif isinstance(node, WaveformPrimitive):
                if waveform is None:
                    waveform = node.__bloqade_ir__()
                else:
                    waveform = waveform.append(node.__bloqade_ir__())
            else:
                break

            curr = curr.next

        return waveform, curr

    def read_drive(self, head) -> ir.Field:
        if head is None:
            return ir.Field({})

        sm = head.node.__bloqade_ir__()
        wf, _ = self.read_waveform(head.next)

        return ir.Field({sm: wf})

    def read_sequence(self) -> ir.Sequence:
        if isinstance(self.stream.curr.node, SequenceBuilder):
            # case with sequence builder object.
            self.sequence = self.stream.read().node._sequence
            return self.sequence

        stream = self.stream.copy()
        while stream.curr is not None:
            coupling_builder, field_builder, spatial_head = self.read_address(stream)

            if coupling_builder is not None:
                # update to new pulse coupling
                self.coupling_name = coupling_builder.__bloqade_ir__()

            if field_builder is not None:
                # update to new field coupling
                self.field_name = field_builder.__bloqade_ir__()

            if spatial_head is None:
                break

            pulse = self.sequence.pulses.get(self.coupling_name, ir.Pulse({}))
            field = pulse.fields.get(self.field_name, ir.Field({}))

            drive = self.read_drive(spatial_head)
            field = field.add(drive)

            pulse.fields[self.field_name] = field
            self.sequence.pulses[self.coupling_name] = pulse

        return self.sequence

    def read_register(self) -> ir.AtomArrangement:
        # register is always head of the stream
        register_node = self.stream.read()
        self.register = register_node.node

        return self.register

    def read_pragmas(self) -> None:
        pragma_types = (
            Assign,
            BatchAssign,
            ListAssign,
            Args,
            Parallelize,
        )

        stream = self.stream.copy()
        curr = stream.read_next(pragma_types)

        while curr is not None:
            node = curr.node

            if isinstance(node, Assign):
                self.static_params = dict(node._static_params)
            elif isinstance(node, BatchAssign) or isinstance(node, ListAssign):
                self.batch_params = node._batch_params
            elif isinstance(node, Args):
                order = node._order

                seen = set()
                dup = []
                for x in order:
                    if x not in seen:
                        seen.add(x)
                    else:
                        dup.append(x)

                if dup:
                    raise ValueError(f"Cannot have duplicate names {dup}.")

                order_names = set([*order])
                vector_names = order_names.intersection(self.vector_node_names)

                if vector_names:
                    raise ValueError(
                        f"Cannot have RunTimeVectors: {vector_names} as an argument."
                    )

                self.order = order

            elif isinstance(node, Parallelize):
                self.register = ir.ParallelRegister(
                    self.register, node._cluster_spacing
                )
            else:
                break

            curr = curr.next

    def parse_register(
        self, builder: Builder
    ) -> Union[ir.AtomArrangement, ir.ParallelRegister]:
        self.reset(builder)
        self.read_register()
        self.read_pragmas()
        return self.register

    def parse_sequence(self, builder: Builder) -> ir.Sequence:
        self.reset(builder)
        self.read_sequence()
        return self.sequence

    def parse_circuit(self, builder: Builder) -> "AnalogCircuit":
        from bloqade.ir.analog_circuit import AnalogCircuit

        self.reset(builder)
        self.read_register()
        self.read_sequence()

        circuit = AnalogCircuit(self.register, self.sequence)

        return circuit

    def parse(self, builder: Builder) -> "Routine":
        from bloqade.ir.analog_circuit import AnalogCircuit
        from bloqade.ir.routine.params import Params
        from bloqade.ir.routine.base import Routine

        self.reset(builder)
        self.read_register()
        self.read_sequence()
        self.read_pragmas()

        params = Params(
            static_params=self.static_params,
            batch_params=self.batch_params,
            args_list=self.order,
        )
        circuit = AnalogCircuit(self.register, self.sequence)

        return Routine(builder, circuit, params)



from unittest.mock import Mock, patch
import pytest
import bloqade.emulate.ir.space
from bloqade.emulate.ir.space import Space
from bloqade.emulate.ir.atom_type import ThreeLevelAtom, TwoLevelAtom
from bloqade.emulate.ir.emulator import Register
import numpy as np


def test_two_level_space():
    positions = [(0, 0), (0, 1)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 2
    assert space.size == 4
    assert np.all(space.configurations == np.array([0, 1, 2, 3]))


def test_two_level_subspace():
    positions = [(0, 0), (0, 1)]
    register = Register(TwoLevelAtom, positions, 1.1)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 2
    assert space.size == 3
    assert np.all(space.configurations == np.array([0, 1, 2]))


def test_two_level_subspace_2():
    positions = [(0, 0), (0, 1), (1, 0), (1, 1)]
    register = Register(TwoLevelAtom, positions, np.sqrt(2) + 0.1)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 4
    assert space.size == 5
    assert np.all(space.configurations == np.array([0, 1, 2, 4, 8]))


def test_two_level_subspace_3():
    space_part = Space.create(Register(TwoLevelAtom, [(0, 0), (0, 1)], 1.1))

    space = Space.create(Register(TwoLevelAtom, [(0, 0), (0, 1), (0, 5), (0, 6)], 1.1))

    config = space_part.configurations
    actual_configs = np.kron(config, np.ones_like(config, dtype=np.uint8)) + np.kron(
        np.ones_like(config) * 2**2, config
    )

    actual_configs.sort()

    np.testing.assert_equal(space.configurations, actual_configs)


def test_three_level_space():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 2
    assert space.size == 9
    assert np.all(space.configurations == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8]))


def test_three_level_subspace():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1.1)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 2
    assert space.size == 8
    assert np.all(space.configurations == np.array([0, 1, 2, 3, 4, 5, 6, 7]))


def test_three_level_subspace_2():
    positions = [(0, 0), (0.5, np.sqrt(3) / 2), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1.1)
    space = Space.create(register)
    print(space)
    assert space.n_atoms == 3
    assert space.size == 20
    assert np.all(
        space.configurations
        == np.array(
            [0, 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 16, 18, 19, 21, 22]
        )
    )


def test_three_level_subspace_3():
    space = Space.create(Register(TwoLevelAtom, [(0, 0), (0, 4), (0, 5)], 1.1))
    assert np.all(space.configurations == np.array([0, 1, 2, 3, 4, 5]))

    space_part = Space.create(Register(ThreeLevelAtom, [(0, 0), (0, 1)], 1.1))

    space = Space.create(
        Register(ThreeLevelAtom, [(0, 0), (0, 1), (0, 5), (0, 6)], 1.1)
    )

    config = space_part.configurations
    actual_configs = np.kron(config, np.ones_like(config, dtype=np.uint8)) + np.kron(
        np.ones_like(config) * 3**2, config
    )

    actual_configs.sort()

    np.testing.assert_equal(space.configurations, actual_configs)


def test_two_level_integer_to_string():
    assert TwoLevelAtom.integer_to_string(0, 2) == "|gg>"
    assert TwoLevelAtom.integer_to_string(1, 2) == "|rg>"
    assert TwoLevelAtom.integer_to_string(2, 2) == "|gr>"
    assert TwoLevelAtom.integer_to_string(3, 2) == "|rr>"


def test_two_level_string_to_integer():
    assert TwoLevelAtom.string_to_integer("|gg>") == 0
    assert TwoLevelAtom.string_to_integer("|rg>") == 1
    assert TwoLevelAtom.string_to_integer("|gr>") == 2
    assert TwoLevelAtom.string_to_integer("|rr>") == 3


def test_three_level_integer_to_string():
    assert ThreeLevelAtom.integer_to_string(0, 2) == "|gg>"
    assert ThreeLevelAtom.integer_to_string(1, 2) == "|hg>"
    assert ThreeLevelAtom.integer_to_string(2, 2) == "|rg>"
    assert ThreeLevelAtom.integer_to_string(3, 2) == "|gh>"
    assert ThreeLevelAtom.integer_to_string(4, 2) == "|hh>"
    assert ThreeLevelAtom.integer_to_string(5, 2) == "|rh>"
    assert ThreeLevelAtom.integer_to_string(6, 2) == "|gr>"
    assert ThreeLevelAtom.integer_to_string(7, 2) == "|hr>"
    assert ThreeLevelAtom.integer_to_string(8, 2) == "|rr>"


def test_three_level_string_to_integer():
    assert ThreeLevelAtom.string_to_integer("|gg>") == 0
    assert ThreeLevelAtom.string_to_integer("|hg>") == 1
    assert ThreeLevelAtom.string_to_integer("|rg>") == 2
    assert ThreeLevelAtom.string_to_integer("|gh>") == 3
    assert ThreeLevelAtom.string_to_integer("|hh>") == 4
    assert ThreeLevelAtom.string_to_integer("|rh>") == 5
    assert ThreeLevelAtom.string_to_integer("|gr>") == 6
    assert ThreeLevelAtom.string_to_integer("|hr>") == 7
    assert ThreeLevelAtom.string_to_integer("|rr>") == 8


def test_two_level_transition_indices():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    # 0. |ggg> -> |rgg>
    # 1. |rgg> ->
    # 2. |grg> -> |rrg>
    # 3. |rrg> ->
    # 4. |ggr> -> |rgr>
    # 5. |rgr> ->
    # 6. |grr> -> |rrr>
    # 7. |rrr> ->

    row_indices, col_indices = space.transition_state_at(0, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, False, True, False, True, False, True, False])
    ).ravel()
    col_indices_expected = np.array([1, 3, 5, 7])

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |grg>
    # 1. |rgg> -> |rrg>
    # 2. |grg> ->
    # 3. |rrg> ->
    # 4. |ggr> -> |grr>
    # 5. |rgr> -> |rrr>
    # 6. |grr> ->
    # 7. |rrr> ->

    row_indices, col_indices = space.transition_state_at(1, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, True, False, False, True, True, False, False])
    ).ravel()
    col_indices_expected = np.array([2, 3, 6, 7])

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |ggr>
    # 1. |rgg> -> |rgr>
    # 2. |grg> -> |grr>
    # 3. |rrg> -> |rrr>
    # 4. |ggr> ->
    # 5. |rgr> ->
    # 6. |grr> ->
    # 7. |rrr> ->

    row_indices, col_indices = space.transition_state_at(2, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, True, True, True, False, False, False, False])
    ).ravel()
    col_indices_expected = np.array([4, 5, 6, 7])

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_two_level_subspace_transition_indices():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    # 0. |ggg> -> |rgg>
    # 1. |rgg> ->
    # 2. |grg> ->
    # 3. |ggr> ->
    # 4. |grr> ->

    row_indices, col_indices = space.transition_state_at(0, 0, 1)

    row_indices_expected = np.array([0])
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|rgg>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |grg>
    # 1. |rgg> ->
    # 2. |grg> ->
    # 3. |ggr> -> |grr>
    # 4. |grr> ->

    row_indices, col_indices = space.transition_state_at(1, 0, 1)

    row_indices_expected = np.array([0, 3])
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|grg>"),
            space.fock_state_to_index("|grr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |ggr>
    # 1. |rgg> ->
    # 2. |grg> -> |grr>
    # 3. |ggr> ->
    # 4. |grr> ->

    row_indices, col_indices = space.transition_state_at(2, 0, 1)

    row_indices_expected = np.array([0, 2])
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|ggr>"),
            space.fock_state_to_index("|grr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_three_level_transition_indices():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    # 0. |gg> -> |hg>
    # 1. |hg> ->
    # 2. |rg> ->
    # 3. |gh> -> |hh>
    # 4. |hh> ->
    # 5. |rh> ->
    # 6. |gr> -> |hr>
    # 7. |hr> ->
    # 8. |rr> ->

    row_indices, col_indices = space.transition_state_at(0, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, False, False, True, False, False, True, False, False])
    ).ravel()
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
            ThreeLevelAtom.string_to_integer("|hr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |gg> ->
    # 1. |hg> -> |rg>
    # 2. |rg> ->
    # 3. |gh> ->
    # 4. |hh> -> |rh>
    # 5. |rh> ->
    # 6. |gr> ->
    # 7. |hr> -> |rr>
    # 8. |rr> ->

    row_indices, col_indices = space.transition_state_at(0, 1, 2)

    row_indices_expected = np.argwhere(
        np.array([False, True, False, False, True, False, False, True, False])
    ).ravel()
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|rg>"),
            ThreeLevelAtom.string_to_integer("|rh>"),
            ThreeLevelAtom.string_to_integer("|rr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_three_level_subspace_transition_indices():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1.1)
    space = Space.create(register)
    print(space)

    # 0. |gg> -> |hg>
    # 1. |hg> ->
    # 2. |rg> ->
    # 3. |gh> -> |hh>
    # 4. |hh> ->
    # 5. |rh> ->
    # 6. |gr> -> |hr>
    # 7. |hr> ->

    row_indices, col_indices = space.transition_state_at(0, 0, 1)

    row_indices_expected = np.array([0, 3, 6])
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
            ThreeLevelAtom.string_to_integer("|hr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |gg> ->
    # 1. |hg> -> |rg>
    # 2. |rg> ->
    # 3. |gh> ->
    # 4. |hh> -> |rh>
    # 5. |rh> ->
    # 6. |gr> ->
    # 7. |hr> ->

    row_indices, col_indices = space.transition_state_at(0, 1, 2)

    row_indices_expected = np.array([1, 4])
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|rg>"),
            ThreeLevelAtom.string_to_integer("|rh>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_two_level_swap_indices():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    row_indices, col_indices = space.swap_state_at(0, 0, 1)

    row_indices_expected = slice(None, None, None)
    col_indices_expected = np.array([1, 0, 3, 2, 5, 4, 7, 6])

    assert row_indices == row_indices_expected
    assert np.all(col_indices == col_indices_expected)

    row_indices, col_indices = space.swap_state_at(1, 0, 1)

    row_indices_expected = slice(None, None, None)
    col_indices_expected = np.array([2, 3, 0, 1, 6, 7, 4, 5])

    assert row_indices == row_indices_expected
    assert np.all(col_indices == col_indices_expected)

    row_indices, col_indices = space.swap_state_at(2, 0, 1)

    row_indices_expected = slice(None, None, None)
    col_indices_expected = np.array([4, 5, 6, 7, 0, 1, 2, 3])

    assert row_indices == row_indices_expected
    assert np.all(col_indices == col_indices_expected)


def test_two_level_subspace_swap_indices():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    # 0. |ggg> -> |rgg>
    # 1. |rgg> -> |ggg>
    # 2. |grg> ->
    # 3. |ggr> ->
    # 4. |grr> ->

    row_indices, col_indices = space.swap_state_at(0, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, True, False, False, False])
    ).ravel()
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|rgg>"),
            space.fock_state_to_index("|ggg>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |grg>
    # 1. |rgg> ->
    # 2. |grg> -> |ggg>
    # 3. |ggr> -> |grr>
    # 4. |grr> -> |ggr>

    row_indices, col_indices = space.swap_state_at(1, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, False, True, True, True])
    ).ravel()
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|grg>"),
            space.fock_state_to_index("|ggg>"),
            space.fock_state_to_index("|grr>"),
            space.fock_state_to_index("|ggr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |ggg> -> |ggr>
    # 1. |rgg> ->
    # 2. |grg> -> |grr>
    # 3. |ggr> -> |ggg>
    # 4. |grr> -> |grg>

    row_indices, col_indices = space.swap_state_at(2, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, False, True, True, True])
    ).ravel()
    col_indices_expected = np.array(
        [
            space.fock_state_to_index("|ggr>"),
            space.fock_state_to_index("|grr>"),
            space.fock_state_to_index("|ggg>"),
            space.fock_state_to_index("|grg>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_three_level_swap_indices():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    # 0. |gg>
    # 1. |hg>
    # 2. |rg>
    # 3. |gh>
    # 4. |hh>
    # 5. |rh>
    # 6. |gr>
    # 7. |hr>
    # 8. |rr>

    row_indices, col_indices = space.swap_state_at(0, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, True, False, True, True, False, True, True, False])
    ).ravel()
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|gg>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
            ThreeLevelAtom.string_to_integer("|gh>"),
            ThreeLevelAtom.string_to_integer("|hr>"),
            ThreeLevelAtom.string_to_integer("|gr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |gg>
    # 1. |hg>
    # 2. |rg>
    # 3. |gh>
    # 4. |hh>
    # 5. |rh>
    # 6. |gr>
    # 7. |hr>
    # 8. |rr>

    row_indices, col_indices = space.swap_state_at(0, 1, 2)

    row_indices_expected = np.argwhere(
        np.array([False, True, True, False, True, True, False, True, True])
    ).ravel()
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|rg>"),
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|rh>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
            ThreeLevelAtom.string_to_integer("|rr>"),
            ThreeLevelAtom.string_to_integer("|hr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_three_level_subspace_swap_indices():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1.1)
    space = Space.create(register)
    print(space)

    # 0. |gg> -> |hg>
    # 1. |hg> -> |gg>
    # 2. |rg> ->
    # 3. |gh> -> |hh>
    # 4. |hh> -> |gh>
    # 5. |rh> ->
    # 6. |gr> -> |hr>
    # 7. |hr> -> |gr>

    row_indices, col_indices = space.swap_state_at(0, 0, 1)

    row_indices_expected = np.argwhere(
        np.array([True, True, False, True, True, False, True, True])
    ).ravel()
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|gg>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
            ThreeLevelAtom.string_to_integer("|gh>"),
            ThreeLevelAtom.string_to_integer("|hr>"),
            ThreeLevelAtom.string_to_integer("|gr>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)

    # 0. |gg> ->
    # 1. |hg> -> |rg>
    # 2. |rg> -> |hg>
    # 3. |gh> ->
    # 4. |hh> -> |rh>
    # 5. |rh> -> |hh>
    # 6. |gr> ->
    # 7. |hr> ->

    row_indices, col_indices = space.swap_state_at(0, 1, 2)

    row_indices_expected = np.array([1, 2, 4, 5])
    col_indices_expected = np.array(
        [
            ThreeLevelAtom.string_to_integer("|rg>"),
            ThreeLevelAtom.string_to_integer("|hg>"),
            ThreeLevelAtom.string_to_integer("|rh>"),
            ThreeLevelAtom.string_to_integer("|hh>"),
        ]
    )

    assert np.all(row_indices == row_indices_expected)
    assert np.all(col_indices == col_indices_expected)


def test_fock_state_to_index():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    assert space.fock_state_to_index("|ggg>") == 0
    assert space.fock_state_to_index("|rgg>") == 1
    assert space.fock_state_to_index("|grg>") == 2
    assert space.fock_state_to_index("|rrg>") == 3
    assert space.fock_state_to_index("|ggr>") == 4
    assert space.fock_state_to_index("|rgr>") == 5
    assert space.fock_state_to_index("|grr>") == 6
    assert space.fock_state_to_index("|rrr>") == 7

    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    # 0. |ggg>
    # 1. |rgg>
    # 2. |grg>
    # 3. |ggr>
    # 4. |grr>

    assert space.fock_state_to_index("|ggg>") == 0
    assert space.fock_state_to_index("|rgg>") == 1
    assert space.fock_state_to_index("|grg>") == 2
    assert space.fock_state_to_index("|ggr>") == 3
    assert space.fock_state_to_index("|grr>") == 4

    with pytest.raises(ValueError):
        space.fock_state_to_index("|rrr>")

    with pytest.raises(ValueError):
        space.fock_state_to_index("|rgr>")


def test_index_to_fock_state():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    assert space.index_to_fock_state(0) == "|ggg>"
    assert space.index_to_fock_state(1) == "|rgg>"
    assert space.index_to_fock_state(2) == "|grg>"
    assert space.index_to_fock_state(3) == "|rrg>"
    assert space.index_to_fock_state(4) == "|ggr>"
    assert space.index_to_fock_state(5) == "|rgr>"
    assert space.index_to_fock_state(6) == "|grr>"
    assert space.index_to_fock_state(7) == "|rrr>"

    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    # 0. |ggg>
    # 1. |rgg>
    # 2. |grg>
    # 3. |ggr>
    # 4. |grr>

    assert space.index_to_fock_state(0) == "|ggg>"
    assert space.index_to_fock_state(1) == "|rgg>"
    assert space.index_to_fock_state(2) == "|grg>"
    assert space.index_to_fock_state(3) == "|ggr>"
    assert space.index_to_fock_state(4) == "|grr>"

    with pytest.raises(ValueError):
        space.index_to_fock_state(-1)

    with pytest.raises(ValueError):
        space.index_to_fock_state(6)


def test_zero_state():
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)
    print(space)

    assert np.all(space.zero_state() == np.array([1, 0, 0, 0, 0, 0, 0, 0]))

    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    assert np.all(space.zero_state() == np.array([1, 0, 0, 0, 0]))


@patch("bloqade.emulate.ir.space.np.random.choice")
def test_sample_state(patch_choice):
    positions = [(0, 0), (0, 1), (1, 0)]
    register = Register(TwoLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    bloqade.emulate.ir.space.np.random.choice.return_value = np.array([0, 1, 2, 3, 4])
    state_vector = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

    expected_bitstrings = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
        ]
    )
    bitstrings = space.sample_state_vector(state_vector, 5)
    print(bitstrings)
    assert np.all(bitstrings == expected_bitstrings)

    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1)
    space = Space.create(register)
    print(space)

    bloqade.emulate.ir.space.np.random.choice.return_value = np.array(
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
    )

    # 0. |gg> -> |gg>
    # 1. |hg> -> |gg>
    # 2. |rg> -> |rg>
    # 3. |gh> -> |gg>
    # 4. |hh> -> |gg>
    # 5. |rh> -> |rg>
    # 6. |gr> -> |gr>
    # 7. |hr> -> |gr>
    # 8. |rr> -> |rr>

    state_vector = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.1, 0.2, 0.3])

    expected_bitstrings = np.array(
        [
            [0, 0],
            [0, 0],
            [1, 0],
            [0, 0],
            [0, 0],
            [1, 0],
            [0, 1],
            [0, 1],
            [1, 1],
        ]
    )

    bitstrings = space.sample_state_vector(state_vector, 9)
    print(bitstrings)

    assert np.all(bitstrings == expected_bitstrings)


def test_str():
    positions = [(0, 0), (0, 1)]
    register = Register(ThreeLevelAtom, positions, 1)
    space = Space.create(register)

    assert (
        str(space) == "0. |gg>\n"
        "1. |hg>\n"
        "2. |rg>\n"
        "3. |gh>\n"
        "4. |hh>\n"
        "5. |rh>\n"
        "6. |gr>\n"
        "7. |hr>\n"
    )

    positions = [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), (3, 0), (3, 1)]
    register = Register(ThreeLevelAtom, positions, 1)
    space = Space.create(register)

    assert (
        str(space) == "   0. |gggggggg>\n"
        "   1. |hggggggg>\n"
        "   2. |rggggggg>\n"
        "   3. |ghgggggg>\n"
        "   4. |hhgggggg>\n"
        "   5. |rhgggggg>\n"
        "   6. |grgggggg>\n"
        "   7. |hrgggggg>\n"
        "   8. |gghggggg>\n"
        "   9. |hghggggg>\n"
        "  10. |rghggggg>\n"
        "  11. |ghhggggg>\n"
        "  12. |hhhggggg>\n"
        "  13. |rhhggggg>\n"
        "  14. |grhggggg>\n"
        "        ...\n"
        "2833. |rhhhrhhr>\n"
        "2834. |grhhrhhr>\n"
        "2835. |hrhhrhhr>\n"
        "2836. |gggrrhhr>\n"
        "2837. |hggrrhhr>\n"
        "2838. |rggrrhhr>\n"
        "2839. |ghgrrhhr>\n"
        "2840. |hhgrrhhr>\n"
        "2841. |rhgrrhhr>\n"
        "2842. |gghrrhhr>\n"
        "2843. |hghrrhhr>\n"
        "2844. |rghrrhhr>\n"
        "2845. |ghhrrhhr>\n"
        "2846. |hhhrrhhr>\n"
        "2847. |rhhrrhhr>\n"
    )


@patch("bloqade.emulate.ir.space.np.iinfo")
def test_dtype(patch_iinfo):
    patch_iinfo.return_value = Mock(max=10)

    positions = [(0, 0), (0, 1)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)

    assert space.index_type == np.int32
    assert space.state_type.char == space.configurations.dtype.char

    positions = [(0, 0), (0, 1), (1, 0), (1, 1)]
    register = Register(TwoLevelAtom, positions, 0)
    space = Space.create(register)

    assert space.index_type == np.int64



from typing import Optional, Union, List
from numbers import Real

from bloqade.builder.parse.trait import Parse, Show

ParamType = Union[Real, List[Real]]


class Builder(Parse, Show):
    def __init__(
        self,
        parent: Optional["Builder"] = None,
    ) -> None:
        self.__parent__ = parent

    def __str__(self):
        return str(self.parse())



from bokeh.plotting import figure
from bokeh.layouts import row, column
from bokeh.models import CustomJS, MultiChoice
from bokeh.models import (
    ColumnDataSource,
    DataCube,
    GroupingInfo,
    StringFormatter,
    TableColumn,
)
from bokeh.models import Tabs, TabPanel, HoverTool, Div, CrosshairTool, Span
from bokeh.palettes import Dark2_5
import itertools
import numpy as np
from typing import List


def mock_data(Ndata):
    out = []
    ch_names = []

    d0 = []
    d1 = []
    scl = []
    unicnt = 0
    for i in range(Ndata):
        x = np.linspace(0, 10, 200)
        y = np.random.normal(size=len(x))
        name = None

        locs = []  # str
        scales = []  # float
        Nloc = np.random.randint(5)
        if Nloc == 0:
            # unifom
            name = "uni[%d]" % (unicnt)
            unicnt += 1
            locs = ["all"]
            scales = [1.0]
        else:
            locs_idx = np.unique(np.random.randint(10, size=Nloc)).flatten()
            locs = ["loc[%d]" % (n) for n in locs_idx]
            scales = np.random.rand(len(locs)).flatten()
            name = "ch[%d]" % (i)

        d0 += [name] * len(locs)
        d1 += list(locs)
        scl += list(scales)

        ch_names.append(name)
        lbl = [name] * len(x)
        out.append(ColumnDataSource(data=dict(wvfm_x=x, wvfm_y=y, lbl=lbl)))

    spmod_source = ColumnDataSource(
        data=dict(
            d0=d0,
            d1=d1,
            px=scl,
        )
    )

    return out, ch_names, spmod_source


def mock_register():
    p = figure(width=300, height=300, toolbar_location="above")
    p.circle(x=np.random.rand(10), y=np.random.rand(10))

    return p


def Field_wvfm(colors, data_sources, ch_names, crx_hair_overlay: List = None):
    p = figure(width=800, height=170)

    # add plots of data:
    glyph_obj = []
    Ndata = len(data_sources)
    for i, color in zip(range(Ndata), colors):
        obj = p.line(
            x="wvfm_x",
            y="wvfm_y",
            source=data_sources[i],
            name="data%d" % (i),
            color=color,
            visible=False,
        )
        glyph_obj.append(obj)

    p.xaxis.axis_label = "Time (us)"

    # add tools:
    # hover:
    hov_tool = HoverTool(mode="vline")
    hov_tool.tooltips = [("(x,y)", "(@wvfm_x, @wvfm_y)"), ("channel", "@lbl")]
    p.add_tools(hov_tool)

    # crosshair:
    if crx_hair_overlay is None:
        crx_hair = CrosshairTool(dimensions="height")
    else:
        crx_hair = CrosshairTool(overlay=crx_hair_overlay)

    p.add_tools(crx_hair)

    # add multi choice for select channel:
    glyph_obj[0].visible = True
    multi_choice = MultiChoice(value=[ch_names[0]], options=ch_names)
    multi_choice.js_on_change(
        "value",
        CustomJS(
            args=dict(glyphobj=glyph_obj, ch_names=ch_names, nelem=len(ch_names)),
            code="""
                                const vals = this.value
                                for(let i=0;i<nelem;i++){
                                    glyphobj[i].visible=false;
                                    for(let j=0;j<vals.length;j++){
                                        if(vals[j]==ch_names[i]){
                                            glyphobj[i].visible=true;
                                            break;
                                        }
                                    }
                                }
                              """,
        ),
    )

    # compose layout:
    return column(multi_choice, p)


def SpacialMod(spmod_data):
    # setting:
    formatter = StringFormatter(font_style="bold")
    target = ColumnDataSource(data=dict(row_indices=[], labels=[]))  # this is needed
    columns = [
        TableColumn(
            field="d1", title="Channel", width=20, sortable=False, formatter=formatter
        ),
        TableColumn(field="px", title="Scale", width=20, sortable=False),
    ]
    grouping = [GroupingInfo(getter="d0")]

    # generate table:
    cube = DataCube(
        source=spmod_data,
        columns=columns,
        grouping=grouping,
        target=target,
        width=200,
        height=170,
    )

    return cube


def assemble_field(left_spmod, right_waveform, title):
    # add div:
    frame = row(left_spmod, right_waveform)
    div = Div(text=title, width=frame.width, height=10)
    return column(div, frame)


def assemble_pulse_panel(field_plots: List, title: str):
    ## link x_range:
    x_range = None
    for field_layout in field_plots:
        wvfm_figure = field_layout.children[1].children[1].children[1]
        if x_range is None:
            x_range = wvfm_figure.x_range
        else:
            wvfm_figure.x_range = x_range

    Panel_Pulse = TabPanel(child=column(*field_plots), title=title)
    return Panel_Pulse  # TabPanel


def assemble_sequences(pulse_tabs: List):
    return Tabs(tabs=pulse_tabs)  # Tab


def format_spmod_data(spmod_ir, **assignments):
    d0 = {"uni": [], "chs": []}
    d1 = {"uni": [], "chs": []}
    scl = {"uni": [], "chs": []}

    locs, scales = spmod_ir._get_data(**assignments)
    name = None
    if str(spmod_ir) == "Uniform":
        name = "Uni"

        d0["uni"] += [name] * len(locs)
        d1["uni"] += list(locs)
        scl["uni"] += list(scales)
    else:
        name = "Channel"

        d0["chs"] += [name] * len(locs)
        d1["chs"] += list(locs)
        scl["chs"] += list(scales)

    return d0, d1, scl


def get_spmod_figure(spmod_ir, **assignments):
    d0, d1, scl = format_spmod_data(spmod_ir, **assignments)
    d0 = d0["uni"] + d0["chs"]
    d1 = d1["uni"] + d1["chs"]
    scl = scl["uni"] + scl["chs"]

    spmod_source = ColumnDataSource(
        data=dict(
            d0=d0,
            d1=d1,
            px=scl,
        )
    )
    cube = SpacialMod(spmod_source)
    return row(cube)


def get_waveform_figure(wvfm_ir, **assignments):
    # Varlist = []
    duration = float(wvfm_ir.duration(**assignments))
    times = np.linspace(0, duration, 1001)
    values = [wvfm_ir.__call__(time, **assignments) for time in times]

    data_source = ColumnDataSource(data=dict(wvfm_x=times, wvfm_y=values))

    fig = figure(
        x_axis_label="Time (us)",
        y_axis_label="Waveform",
        aspect_ratio=(800 / 170),
        sizing_mode="scale_both",
    )

    fig.line(
        x="wvfm_x",
        y="wvfm_y",
        source=data_source,
        line_width=2,
    )
    hov_tool = HoverTool(mode="vline")
    hov_tool.tooltips = [("(x,y)", "(@wvfm_x, @wvfm_y)")]

    crx_hair = CrosshairTool(dimensions="height")

    fig.add_tools(hov_tool)
    fig.add_tools(crx_hair)

    return fig


## =====================================================
# below are formatting IR data, and called by IR.
def format_field_data(field_ir, **assignments):
    # return should be
    # data: List[ColumnDataSources],
    # List[str]:ch_names,
    # spmod_source:ColumnDataSources

    out = {"uni": [], "chs": []}
    d0 = {"uni": [], "chs": []}
    d1 = {"uni": [], "chs": []}
    scl = {"uni": [], "chs": []}
    ch_names = {"uni": [], "chs": []}
    for spmod_ir, wvfm_ir in field_ir.drives.items():
        ## deal with spmod:
        locs, scales = spmod_ir._get_data(**assignments)

        ## deal with wvfm:
        times, values = wvfm_ir._get_data(1000, **assignments)

        if str(spmod_ir) == "Uniform":
            name = "Uni[%d]" % (len(ch_names["uni"]))
            ch_names["uni"].append(name)

            lbl = [name] * len(times)
            out["uni"].append(
                ColumnDataSource(data=dict(wvfm_x=times, wvfm_y=values, lbl=lbl))
            )

            d0["uni"] += [name] * len(locs)
            d1["uni"] += list(locs)
            scl["uni"] += list(scales)
        else:
            name = "Ch[%d]" % (len(ch_names["chs"]))
            ch_names["chs"].append(name)

            lbl = [name] * len(times)
            out["chs"].append(
                ColumnDataSource(data=dict(wvfm_x=times, wvfm_y=values, lbl=lbl))
            )

            d0["chs"] += [name] * len(locs)
            d1["chs"] += list(locs)
            scl["chs"] += list(scales)

    ch_names = ch_names["uni"] + ch_names["chs"]
    out = out["uni"] + out["chs"]

    d0 = d0["uni"] + d0["chs"]
    d1 = d1["uni"] + d1["chs"]
    scl = scl["uni"] + scl["chs"]

    spmod_source = ColumnDataSource(
        data=dict(
            d0=d0,
            d1=d1,
            px=scl,
        )
    )

    return out, ch_names, spmod_source


def get_field_figure(ir_field, title, indicator, **assignments):
    dats, names, spinfo = format_field_data(ir_field, **assignments)
    if indicator is None:
        fig = Field_wvfm(
            colors=itertools.cycle(Dark2_5), data_sources=dats, ch_names=names
        )
    else:
        fig = Field_wvfm(
            colors=itertools.cycle(Dark2_5),
            data_sources=dats,
            ch_names=names,
            crx_hair_overlay=indicator,
        )
    cube = SpacialMod(spinfo)
    p1 = assemble_field(cube, fig, title)

    return p1


def get_pulse_panel(ir_pulse, title: str = None, **assignments):
    pulse_name, fields = ir_pulse._get_data(**assignments)
    if pulse_name is None:
        if title is not None:
            pulse_name = title
        else:
            pulse_name = "Pulse"

    shared_indicator = Span(dimension="height")
    pulse_figs = []
    for Field_name, field in fields.items():
        p = get_field_figure(field, str(Field_name), shared_indicator, **assignments)
        pulse_figs.append(p)
    p_all = assemble_pulse_panel(pulse_figs, pulse_name)
    return p_all


def get_pulse_figure(ir_pulse, title: str = None, **assginments):
    return Tabs(tabs=[get_pulse_panel(ir_pulse, title, **assginments)])


def get_sequence_figure(ir_seq, **assignments):
    seq_name, pulses = ir_seq._get_data(**assignments)

    pulse_all = []
    for lvl, pulse in pulses.items():
        if seq_name is None:
            p = get_pulse_panel(pulse, title=str(lvl), **assignments)
        else:
            p = get_pulse_panel(pulse, title=str(lvl) + f"({seq_name})", **assignments)
        pulse_all.append(p)

    Seq = assemble_sequences(pulse_all)
    return Seq


def get_ir_figure(ir, **assignments):
    # sequence, spmod and waveform does not need extra
    # pulse, field need extra arguments
    # so we handle separately
    from bloqade.ir.control.sequence import SequenceExpr
    from bloqade.ir.control.field import SpatialModulation
    from bloqade.ir.control.waveform import Waveform

    if isinstance(ir, SequenceExpr):
        return get_sequence_figure(ir, **assignments)
    elif isinstance(ir, SpatialModulation):
        return get_spmod_figure(ir, **assignments)
    elif isinstance(ir, Waveform):
        return get_waveform_figure(ir, **assignments)
    else:
        raise NotImplementedError(f"not supported IR for display, got {type(ir)}")



from bloqade import start, var
import pytest


def test_options_1():
    def detuning(t, x, y):
        return y * t + x

    (
        start.add_position((0, 0))
        .add_position((0, "d"))
        .rydberg.rabi.amplitude.uniform.fn(detuning, 4)
        .sample(0.05)
        .assign(x=1)
        .batch_assign(d=[0, 1])
        .args([var("y")])
        .parallelize(20)
        .quera.mock()
        ._compile(100, args=(2,))
    )


def test_options_2():
    def detuning(t, x, y):
        return y * t + x

    (
        start.add_position((0, 0))
        .add_position((0, "d"))
        .rydberg.rabi.amplitude.uniform.fn(detuning, 4)
        .sample(0.05)
        .assign(x=1)
        .args(["y", "d"])
        .parallelize(20)
        .quera.mock()
        ._compile(100, args=(2, 4))
    )


def test_options_3():
    def detuning(t, x, y):
        return y * t + x

    (
        start.add_position((0, 0))
        .add_position((0, "d"))
        .rydberg.rabi.amplitude.uniform.fn(detuning, 4)
        .sample(0.05)
        .assign(x=1)
        .batch_assign(d=[0, 1], y=[0, 1])
        .parallelize(20)
        .quera.mock()
        ._compile(100)
    )


def test_options_4():
    def detuning(t, x, y):
        return y * t + x

    program = (
        start.add_position((0, 0))
        .add_position((0, "d"))
        .rydberg.rabi.amplitude.uniform.fn(detuning, 4)
        .sample(0.05)
        .assign(x=1)
        .batch_assign(d=[0, 1], y=[0, 1])
        .quera.mock()
    )

    program._compile(100)

    with pytest.raises(ValueError):
        # checking that using `arg` is not allowed without args
        program._compile(100, args=(2,))



from bloqade.emulate.ir.emulator import EmulatorProgram
from bloqade.emulate.ir.space import Space
from bloqade.emulate.sparse_operator import (
    IndexMapping,
    SparseMatrixCSC,
    SparseMatrixCSR,
)
from dataclasses import dataclass, field
from numpy.typing import NDArray
from beartype.typing import List, Callable, Union, Optional
from beartype.vale import IsAttr, IsEqual
from typing import Annotated
from beartype import beartype
import numpy as np
from scipy.integrate import ode


SparseOperator = Union[IndexMapping, SparseMatrixCSR, SparseMatrixCSC]


@dataclass(frozen=True)
class DetuningOperator:
    diagonal: NDArray
    amplitude: Optional[Callable[[float], float]] = None

    def get_diagonal(self, time: float):
        if self.amplitude:
            return self.diagonal * self.amplitude(time)
        else:
            return self.diagonal


@dataclass(frozen=True)
class RabiOperator:
    op: SparseOperator
    amplitude: Callable[[float], float]
    phase: Optional[Callable[[float], float]] = None

    def dot(self, register: NDArray, output: NDArray, time: float):
        amplitude = self.amplitude(time) / 2
        if self.phase is None:
            return self.op.matvec(register, out=output, scale=amplitude)

        amplitude *= np.exp(1j * self.phase(time))
        self.op.matvec(register, out=output, scale=amplitude)
        self.op.T.matvec(register, out=output, scale=np.conj(amplitude))

        return output


@dataclass(frozen=True)
class RydbergHamiltonian:
    emulator_ir: EmulatorProgram
    space: Space
    rydberg: NDArray
    detuning_ops: List[DetuningOperator] = field(default_factory=list)
    rabi_ops: List[RabiOperator] = field(default_factory=list)

    def _ode_complex_kernel(self, time: float, register: NDArray, output: NDArray):
        diagonal = sum(
            (detuning.get_diagonal(time) for detuning in self.detuning_ops),
            start=self.rydberg,
        )

        np.multiply(diagonal, register, out=output)
        for rabi_op in self.rabi_ops:
            rabi_op.dot(register, output, time)

        output *= -1j
        return output

    def _ode_real_kernel(self, time: float, register: NDArray, output: NDArray):
        # this is needed to use solver that only work on real-valued states
        return self._ode_complex_kernel(
            time, register.view(np.complex128), output
        ).view(np.float64)

    def _ode_complex_kernel_int(self, time: float, register: NDArray, output: NDArray):
        diagonal = sum(
            (detuning.get_diagonal(time) for detuning in self.detuning_ops),
        )

        u = np.exp(-1j * self.rydberg * time)

        int_register = u * register

        np.multiply(diagonal, int_register, out=output)
        for rabi_op in self.rabi_ops:
            rabi_op.dot(int_register, output, time)

        np.conj(u, out=u)
        np.multiply(u, output, out=output)

        output *= -1j
        return output

    def _ode_real_kernel_int(self, time: float, register: NDArray, output: NDArray):
        # this is needed to use solver that only work on real-valued states
        return self._ode_complex_kernel_int(
            time, register.view(np.complex128), output
        ).view(np.float64)


RealArray = Annotated[NDArray[np.floating], IsAttr["ndim", IsEqual[1]]]
Complexarray = Annotated[NDArray[np.complexfloating], IsAttr["ndim", IsEqual[1]]]
StateArray = Union[RealArray, Complexarray]


@dataclass(frozen=True)
class AnalogGate:
    SUPPORTED_SOLVERS = ["lsoda", "dop853", "dopri5"]

    hamiltonian: RydbergHamiltonian

    @staticmethod
    def _error_check_dop(status_code: int):
        if status_code in [1, 2]:
            return
        elif status_code == -1:
            raise RuntimeError("DOP853/DOPRI5: Input is not consistent.")
        elif status_code == -2:
            raise RuntimeError("DOP853/DOPRI5: Larger nsteps is needed.")
        elif status_code == -3:
            raise RuntimeError("DOP853/DOPRI5: Step size becomes too small.")
        elif status_code == -4:
            raise RuntimeError(
                "DOP853/DOPRI5: Problem is probably stiff (interrupted)."
            )
        else:
            raise RuntimeError(f"DOP853/DOPRI5: unhandled status code {status_code}")

    @staticmethod
    def _error_check_lsoda(status_code: int):
        if status_code == 2:
            return
        elif status_code == -1:
            raise RuntimeError(
                "LSODA: Excess work done on this call (perhaps wrong Dfun type)."
            )
        elif status_code == -2:
            raise RuntimeError(
                "LSODA: Excess accuracy requested (tolerances too small)."
            )
        elif status_code == -3:
            raise RuntimeError("LSODA: Illegal input detected (internal error).")
        elif status_code == -4:
            raise RuntimeError("LSODA: Repeated error test failures (internal error).")
        elif status_code == -5:
            raise RuntimeError(
                "LSODA: Repeated convergence failures "
                "(perhaps bad Jacobian or tolerances)."
            )
        elif status_code == -6:
            raise RuntimeError("LSODA: Error weight became zero during problem.")
        elif status_code == -7:
            raise RuntimeError(
                "LSODA: Internal workspace insufficient to finish (internal error)."
            )
        else:
            raise RuntimeError(f"LSODA: unhandled status code {status_code}")

    @staticmethod
    def _error_check(solver_name: str, status_code: int):
        if solver_name == "lsoda":
            AnalogGate._error_check_lsoda(status_code)
        elif solver_name in ["dop853", "dopri5"]:
            AnalogGate._error_check_dop(status_code)

    def _apply(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
    ):
        if state is None:
            state = self.hamiltonian.space.zero_state()

        if solver_name not in AnalogGate.SUPPORTED_SOLVERS:
            raise ValueError(f"'{solver_name}' not supported.")

        duration = self.hamiltonian.emulator_ir.duration

        state = np.asarray(state).astype(np.complex128, copy=False)

        solver = ode(self.hamiltonian._ode_real_kernel)
        solver.set_f_params(np.zeros_like(state, dtype=np.complex128))
        solver.set_initial_value(state.view(np.float64))
        solver.set_integrator(solver_name, atol=atol, rtol=rtol, nsteps=nsteps)

        if any(time >= duration or time < 0.0 for time in times):
            raise ValueError("Times must be between 0 and duration.")

        times = [*times, duration]

        for time in times:
            if time == 0.0:
                yield state
                continue
            solver.integrate(time)
            AnalogGate._error_check(solver_name, solver.get_return_code())
            yield solver.y.view(np.complex128)

    def _apply_interation_picture(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
    ):
        if state is None:
            state = self.hamiltonian.space.zero_state()

        if solver_name not in AnalogGate.SUPPORTED_SOLVERS:
            raise ValueError(f"'{solver_name}' not supported.")

        duration = self.hamiltonian.emulator_ir.duration

        state = np.asarray(state).astype(np.complex128, copy=False)

        solver = ode(self.hamiltonian._ode_real_kernel_int)
        solver.set_f_params(np.zeros_like(state, dtype=np.complex128))
        solver.set_initial_value(state.view(np.float64))
        solver.set_integrator(solver_name, atol=atol, rtol=rtol, nsteps=nsteps)

        if any(time >= duration or time < 0.0 for time in times):
            raise ValueError("Times must be between 0 and duration.")

        times = [*times, duration]

        for time in times:
            if time == 0.0:
                yield state
                continue
            solver.integrate(time)
            AnalogGate._error_check(solver_name, solver.get_return_code())
            u = np.exp(-1j * time * self.hamiltonian.rydberg)
            yield u * solver.y.view(np.complex128)

    @beartype
    def apply(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
        interaction_picture: bool = False,
    ):
        if interaction_picture:
            return self._apply_interation_picture(
                state,
                solver_name=solver_name,
                atol=atol,
                rtol=rtol,
                nsteps=nsteps,
                times=times,
            )
        else:
            return self._apply(
                state,
                solver_name=solver_name,
                atol=atol,
                rtol=rtol,
                nsteps=nsteps,
                times=times,
            )

    @beartype
    def run(
        self,
        shots: int = 1,
        solver_name: str = "dop853",
        atol: float = 1e-14,
        rtol: float = 1e-7,
        nsteps: int = 2_147_483_647,
        interaction_picture: bool = False,
        project_hyperfine: bool = True,
    ):
        """Run the emulation with all atoms in the ground state,
        sampling the final state vector."""

        options = dict(
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )

        state = self.hamiltonian.space.zero_state()
        (result,) = self.apply(state, **options)
        result /= np.linalg.norm(result)

        return self.hamiltonian.space.sample_state_vector(
            result, shots, project_hyperfine=project_hyperfine
        )



from beartype.typing import TYPE_CHECKING
from bloqade.builder.base import Builder


if TYPE_CHECKING:
    from bloqade.ir.routine.braket import (
        BraketHardwareRoutine,
        BraketLocalEmulatorRoutine,
    )


class BraketService(Builder):
    @property
    def braket(self):
        """
        Specify the Braket backend. This allows you to access the AWS Braket local
        emulator OR go submit things to QuEra hardware on AWS Braket service.

        - Possible Next Steps are:
            - `...braket.aquila()`: target submission to the QuEra Aquila QPU
            - `...braket.local_emulator()`: target submission to the Braket
            local emulator
        """
        return BraketDeviceRoute(self)


class BraketDeviceRoute(Builder):
    def device(self, device_arn) -> "BraketHardwareRoutine":
        """
        Specify QPU based on the device ARN on Braket to submit your program to.

        The number of shots you specify in the subsequent `.run` method will either:
            - dictate the number of times your program is run
            - dictate the number of times per parameter your program is run if
                you have a variable with batch assignments/intend to conduct
                a parameter sweep


        - Possible next steps are:
            - `...device(arn).run(shots)`: To submit to hardware and WAIT for
                results (blocking)
            - `...device(arn).run_async(shots)`: To submit to hardware and immediately
                allow for other operations to occur
        """
        return self.parse().braket.device(device_arn)

    def aquila(self) -> "BraketHardwareRoutine":
        """
        Specify QuEra's Aquila QPU on Braket to submit your program to.

        The number of shots you specify in the subsequent `.run` method will either:
            - dictate the number of times your program is run
            - dictate the number of times per parameter your program is run if
              you have a variable with batch assignments/intend to conduct
              a parameter sweep


        - Possible next steps are:
            - `...aquila().run(shots)`: To submit to hardware and WAIT for
                results (blocking)
            - `...aquila().run_async(shots)`: To submit to hardware and immediately
                allow for other operations to occur
        """
        return self.parse().braket.aquila()

    def local_emulator(self) -> "BraketLocalEmulatorRoutine":
        """
        Specify the Braket local emulator to submit your program to.

        - The number of shots you specify in the subsequent `.run` method will either:
            - dictate the number of times your program is run
            - dictate the number of times per parameter your program is run if
              you have a variable with batch assignments/intend to
              conduct a parameter sweep
        - Possible next steps are:
            - `...local_emulator().run(shots)`: to submit to the emulator
                and await results

        """
        return self.parse().braket.local_emulator()



from bloqade.builder.drive import Drive
from bloqade.builder.coupling import LevelCoupling
from bloqade.builder.field import Field, Rabi
from bloqade.builder.pragmas import (
    Assignable,
    BatchAssignable,
    Parallelizable,
    AddArgs,
)
from bloqade.builder.backend import BackendRoute


class PulseRoute(Drive, LevelCoupling, Field, Rabi):
    pass


class PragmaRoute(Assignable, BatchAssignable, Parallelizable, AddArgs, BackendRoute):
    pass


class WaveformRoute(PulseRoute, PragmaRoute):
    pass



import pytest

import bloqade.submission.quera
import bloqade.submission.ir.task_specification as task_spec
from bloqade.submission.ir.task_results import QuEraTaskResults, QuEraTaskStatusCode
from bloqade.submission.base import ValidationError
from bloqade.submission.capabilities import get_capabilities
from bloqade.submission.ir.capabilities import QuEraCapabilities

from unittest.mock import patch


def get_task_ir():
    return task_spec.QuEraTaskSpecification(
        nshots=10,
        lattice=task_spec.Lattice(sites=[(0, 0)], filling=[1]),
        effective_hamiltonian=task_spec.EffectiveHamiltonian(
            rydberg=task_spec.RydbergHamiltonian(
                rabi_frequency_amplitude=task_spec.RabiFrequencyAmplitude(
                    global_=task_spec.GlobalField(
                        times=[0, 1e-6, 2e-6, 3e-6, 4e-6],
                        values=[0, 15e6, 15e6, 0],
                    )
                ),
                rabi_frequency_phase=task_spec.RabiFrequencyPhase(
                    global_=task_spec.GlobalField(
                        times=[0, 4e-6],
                        values=[0, 0],
                    )
                ),
                detuning=task_spec.Detuning(
                    global_=task_spec.GlobalField(
                        times=[0, 1e-6, 2e-6, 3e-6, 4e-6],
                        values=[0, 15e6, 15e6, 0],
                    )
                ),
            )
        ),
    )


def mock_results():
    return {
        "task_status": "Completed",
        "shot_outputs": [
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
        ],
    }


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_submit(*args):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    queue = bloqade.submission.quera.QueueApi(**api_config)

    queue.post_task.return_value = "task_id"

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.submit_task(get_task_ir()) == "task_id"
    queue.post_task.assert_called_once()


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_validate_fail(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    bloqade.submission.quera.QueueApi.ValidationError = RuntimeError
    queue = bloqade.submission.quera.QueueApi(**api_config)

    queue.validate_task.side_effect = RuntimeError("error")

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    with pytest.raises(ValidationError):
        backend.validate_task(get_task_ir())

    queue.validate_task.assert_called_once()


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_validate_pass(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    queue = bloqade.submission.quera.QueueApi(**api_config)

    queue.validate_task.return_value = None

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.validate_task(get_task_ir()) is None

    queue.validate_task.assert_called_once()


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_cancel(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    queue = bloqade.submission.quera.QueueApi(**api_config)

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    backend.cancel_task("task_id")

    queue.cancel_task_in_queue.assert_called_once_with("task_id")


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_status(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    queue = bloqade.submission.quera.QueueApi(**api_config)

    queue.get_task_status_in_queue.return_value = "Completed"

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.task_status("task_id") == QuEraTaskStatusCode.Completed

    queue.get_task_status_in_queue.assert_called_once_with("task_id")


@patch("bloqade.submission.quera.QueueApi")
def teest_quera_backend_task_results(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )

    queue = bloqade.submission.quera.QueueApi(**api_config)

    queue.poll_task_results.return_value = mock_results()

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.task_results("task_id") == QuEraTaskResults(**mock_results())

    queue.poll_task_results.assert_called_once_with("task_id")


@patch("bloqade.submission.quera.QueueApi")
def test_quera_backend_get_capabilities(QueueApi):
    api_config = dict(
        api_hostname="https://api.que-ee.com", qpu_id="qpu-1", api_stage="v0"
    )
    capabilities_dict = get_capabilities().dict()

    queue = bloqade.submission.quera.QueueApi(**api_config)
    queue.get_capabilities.return_value = capabilities_dict

    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.get_capabilities() == QuEraCapabilities(**capabilities_dict)
    queue.get_capabilities.assert_called_once()

    queue.reset_mock()

    queue.get_capabilities.return_value = Exception("error")
    backend = bloqade.submission.quera.QuEraBackend(**api_config)

    assert backend.get_capabilities() == QuEraCapabilities(**get_capabilities().dict())
    queue.get_capabilities.assert_called_once()


test_quera_backend_validate_fail()
test_quera_backend_validate_pass()



from bloqade.ir.location import ListOfLocations
import tempfile

prog = (
    ListOfLocations()
    .add_position((0, 0))
    .add_position((0, 6))
    .add_position((0, 3))
    .rydberg.detuning.uniform.piecewise_linear(
        durations=["up_time", "anneal_time", "up_time"],
        values=[
            "initial_detuning",
            "initial_detuning",
            "final_detuning",
            "final_detuning",
        ],
    )
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        durations=["up_time", "anneal_time", "up_time"],
        values=[0, "rabi_amplitude_max", "rabi_amplitude_max", 0],
    )
    .assign(
        up_time=0.1,
        anneal_time=3.8,
        initial_detuning=-15,
        final_detuning=10,
        rabi_amplitude_max=15,
    )
)


with tempfile.NamedTemporaryFile() as f:
    future = prog.quera.mock(state_file=f.name).run_async(shots=100)
    future.pull()
    future2 = future.remove_tasks("Completed")
    future2

print(future.report().bitstrings)
print(future.report().counts)
print(future.report().dataframe)
print(future.report().rydberg_densities())



from bloqade.ir.control.waveform import (
    Waveform,
    AlignedWaveform,
    Linear,
    Constant,
    Poly,
    Smooth,
    Slice,
    Append,
    Negative,
    Scale,
    Add,
    Record,
    PythonFn,
    Sample,
)
from typing import Any


class WaveformVisitor:
    def visit_alligned(self, ast: AlignedWaveform) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for AlignedWaveform."
        )

    def visit_linear(self, ast: Linear) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Linear."
        )

    def visit_constant(self, ast: Constant) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Constant."
        )

    def visit_poly(self, ast: Poly) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Poly."
        )

    def visit_smooth(self, ast: Smooth) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Smooth."
        )

    def visit_slice(self, ast: Slice) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Slice."
        )

    def visit_append(self, ast: Append) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Append."
        )

    def visit_negative(self, ast: Negative) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Negative."
        )

    def visit_scale(self, ast: Scale) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Scale."
        )

    def visit_add(self, ast: Add) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Add."
        )

    def visit_record(self, ast: Record) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Record."
        )

    def visit_python_fn(self, ast: PythonFn) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for PythonFn."
        )

    def visit_sample(self, ast: Sample) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Sample."
        )

    def visit(self, ast: Waveform):
        if isinstance(ast, Linear):
            return self.visit_linear(ast)
        elif isinstance(ast, Constant):
            return self.visit_constant(ast)
        elif isinstance(ast, PythonFn):
            return self.visit_python_fn(ast)
        elif isinstance(ast, Poly):
            return self.visit_poly(ast)
        elif isinstance(ast, Append):
            return self.visit_append(ast)
        elif isinstance(ast, Record):
            return self.visit_record(ast)
        elif isinstance(ast, Sample):
            return self.visit_sample(ast)
        elif isinstance(ast, Slice):
            return self.visit_slice(ast)
        elif isinstance(ast, Negative):
            return self.visit_negative(ast)
        elif isinstance(ast, Scale):
            return self.visit_scale(ast)
        elif isinstance(ast, Add):
            return self.visit_add(ast)
        elif isinstance(ast, Smooth):
            return self.visit_smooth(ast)
        elif isinstance(ast, AlignedWaveform):
            return self.visit_alligned(ast)
        else:
            raise ValueError(f"{ast.__class__} is not a Waveform AST type.")



# import bloqade.lattice as lattice

# geo = lattice.Square(3)
# prog = geo.rydberg.detuning
# for i in range(5):
#     prog = prog.location(i)
# prog.linear(start=1.0, stop=2.0, duration="x")
# import pytest
import bloqade.ir as ir
from bloqade.builder import waveform

# import bloqade.builder.backend as builder_backend
import bloqade.ir.routine.quera as quera
import bloqade.ir.routine.braket as braket
from plum import NotFoundLookupError

from bloqade.ir.control.waveform import to_waveform
from bloqade.ir import rydberg, detuning, hyperfine, rabi
from bloqade import start, cast, var

# from bloqade.ir.location import Square, Chain
import numpy as np
import pytest


def test_assign_checks():
    t = var("t")

    t_2 = var("T").max(t)
    t_1 = var("T").min(t)

    delta = var("delta") / (2 * np.pi)
    omega_max = var("omega_max") * 2 * np.pi

    @to_waveform(t_2)
    def detuning(t, u):
        return np.abs(t) * u

    with pytest.raises(ValueError):
        (
            start.add_position(("x", "y"))
            .rydberg.rabi.amplitude.scale("mask")
            .piecewise_linear([0.1, t - 0.2, 0.1], [0, omega_max, omega_max, 0])
            .slice(t_1, t_2)
            .uniform.poly([1, 2, 3, 4], t_1)
            .detuning.uniform.constant(10, t_2)
            .uniform.linear(0, delta, t_1)
            .phase.uniform.apply(-(2 * detuning))
            .assign(c=10)
        )

    with pytest.raises(ValueError):
        (
            start.add_position(("x", "y"))
            .rydberg.rabi.amplitude.scale("mask")
            .piecewise_linear([0.1, t - 0.2, 0.1], [0, omega_max, omega_max, 0])
            .slice(t_1, t_2)
            .uniform.poly([1, 2, 3, 4], t_1)
            .detuning.uniform.constant(10, t_2)
            .uniform.linear(0, delta, t_1)
            .phase.uniform.apply(-(2 * detuning))
            .batch_assign(c=[10])
        )


def test_add_position_dispatch():
    position = np.array([[1, 2], [3, 4]])
    position_list = list(map(tuple, position.tolist()))

    a = start.add_position(position, np.array([True, False]))
    b = start.add_position(position_list, [True, False])
    c = start.add_position(position_list[0]).add_position(position_list[1], False)

    assert a.location_list == b.location_list
    assert a.location_list == c.location_list

    with pytest.raises(AssertionError):
        start.add_position(position_list, [True])

    with pytest.raises(NotFoundLookupError):
        start.add_position(position_list, True)

    with pytest.raises(NotFoundLookupError):
        start.add_position(position_list, np.array([True, True]))

    with pytest.raises(NotFoundLookupError):
        start.add_position(position, [True, True])


def test_piecewise_const():
    prog = start.rydberg.detuning.uniform.piecewise_constant(
        durations=[0.1, 3.1, 0.05], values=[4, 4, 7.5]
    )

    ## inspect ir
    node1 = prog.__bloqade_ir__()
    ir1 = node1.waveforms[2]
    assert ir1.value == cast(7.5)
    assert ir1.duration == cast(0.05)

    ir2 = node1.waveforms[1]
    assert ir2.value == cast(4)
    assert ir2.duration == cast(3.1)

    ir3 = node1.waveforms[0]
    assert ir3.value == cast(4)
    assert ir3.duration == cast(0.1)


def test_registers():
    waveform = (
        ir.Linear("initial_detuning", "initial_detuning", "up_time")
        .append(ir.Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(ir.Linear("final_detuning", "final_detuning", "up_time"))
    )
    prog1 = start.rydberg.detuning.uniform.apply(waveform)
    reg = prog1.parse_register()

    assert reg.n_atoms == 0
    assert reg.n_dims is None


def test_scale():
    prog = start
    prog = prog.rydberg.detuning.location(1, 1.2).piecewise_linear(
        [0.1, 3.8, 0.1], [-10, -10, "a", "b"]
    )

    ## let Emit build ast
    seq = prog.parse_sequence()

    # print(type(list(seq.pulses.keys())[0]))
    Loc1 = list(seq.pulses[rydberg].fields[detuning].drives.keys())[0]

    assert type(Loc1) == ir.ScaledLocations
    assert Loc1.value[ir.Location(1)] == cast(1.2)


def test_scale_location():
    prog = start.rydberg.detuning.location([1, 2], [1.2, 3.3])

    assert prog._scaled_locations == {1: cast(1.2), 2: cast(3.3)}


def test_build_ast_Scale():
    prog = start.rydberg.detuning.location([1, 2], [1.2, 3.3]).piecewise_constant(
        durations=[0.1], values=[1]
    )

    # compile ast:
    tmp = prog.parse_sequence()

    locs = list(tmp.pulses[rydberg].fields[detuning].drives.keys())[0]
    wvfm = tmp.pulses[rydberg].fields[detuning].drives[locs]

    assert locs == ir.ScaledLocations(
        {ir.Location(2): cast(3.3), ir.Location(1): cast(1.2)}
    )
    assert wvfm == ir.Constant(value=cast(1), duration=cast(0.1))


def test_spatial_var():
    prog = start.rydberg.detuning.scale("a")

    assert prog._name_or_list == "a"

    prog = prog.piecewise_constant([0.1], [30])

    # test build ast:
    seq = prog.parse_sequence()

    assert seq.pulses[rydberg].fields[detuning].drives[
        ir.RunTimeVector("a")
    ] == ir.Constant(value=30, duration=0.1)


def test_issue_107():
    waveform = (
        ir.Linear("initial_detuning", "initial_detuning", "up_time")
        .append(ir.Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(ir.Linear("final_detuning", "final_detuning", "up_time"))
    )

    prog1 = start.rydberg.detuning.uniform.apply(waveform)
    prog2 = start.rydberg.detuning.uniform.piecewise_linear(
        durations=["up_time", "anneal_time", "up_time"],
        values=[
            "initial_detuning",
            "initial_detuning",
            "final_detuning",
            "final_detuning",
        ],
    )

    assert prog1.parse_sequence() == prog2.parse_sequence()


def test_issue_150():
    prog = start.rydberg.detuning.uniform.linear(0, 1, 1).amplitude.uniform.linear(
        0, 2, 1
    )

    assert prog.parse_sequence() == ir.Sequence(
        {
            ir.rydberg: ir.Pulse(
                {
                    ir.detuning: ir.Field({ir.Uniform: ir.Linear(0, 1, 1)}),
                    ir.rabi.amplitude: ir.Field({ir.Uniform: ir.Linear(0, 2, 1)}),
                }
            )
        }
    )


def test_303_replicate_channel_should_add():
    prog = (
        start.rydberg.detuning.uniform.linear(0, 1, 1)
        .rabi.amplitude.uniform.linear(1, 2, 1)
        .detuning.uniform.linear(0, 2, 3)
    )

    assert prog.parse_sequence() == ir.Sequence(
        {
            ir.rydberg: ir.Pulse(
                {
                    ir.detuning: ir.Field(
                        {ir.Uniform: ir.Linear(0, 1, 1) + ir.Linear(0, 2, 3)}
                    ),
                    ir.rabi.amplitude: ir.Field({ir.Uniform: ir.Linear(1, 2, 1)}),
                }
            )
        }
    )

    prog1 = (
        start.rydberg.detuning.uniform.linear(0, 1, 1)
        .rabi.amplitude.uniform.linear(1, 2, 1)
        .rydberg.detuning.uniform.linear(0, 2, 3)
    )

    assert prog1.parse_sequence() == prog.parse_sequence()


def test_record():
    prog = start
    prog = (
        prog.rydberg.detuning.location(1)
        .piecewise_constant([0.1], [30])
        .record("detuning")
    )

    assert type(prog) == waveform.Record

    seq = prog.parse_sequence()
    assert seq.pulses[rydberg].fields[detuning].drives[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Record(waveform=ir.Constant(value=30, duration=0.1), var=cast("detuning"))


def test_hyperfine_phase():
    prog = start.hyperfine.rabi.phase.location(1).piecewise_constant([0.1], [30])

    seq = prog.parse_sequence()

    assert seq.pulses[hyperfine].fields[rabi.phase].drives[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Constant(value=30, duration=0.1)


def test_hyperfine_amplitude():
    prog = start.hyperfine.rabi.amplitude.location(1).piecewise_constant([0.1], [30])

    seq = prog.parse_sequence()

    assert seq.pulses[hyperfine].fields[rabi.amplitude].drives[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Constant(value=30, duration=0.1)


def test_piecewise_constant_mismatch():
    with pytest.raises(AssertionError):
        start.hyperfine.rabi.amplitude.location(1).piecewise_constant([0.1, 0.5], [30])


def test_piecewise_linear_mismatch():
    with pytest.raises(AssertionError):
        start.hyperfine.rabi.amplitude.location(1).piecewise_linear(
            durations=[0.1, 0.5], values=[30, 20]
        )


def test_backend_route():
    prog = start.rydberg.detuning.uniform.constant(4, 4)

    assert isinstance(prog.device("quera.aquila"), quera.QuEraHardwareRoutine)
    assert isinstance(prog.device("braket.aquila"), braket.BraketHardwareRoutine)
    assert isinstance(
        prog.device("braket.local_emulator"), braket.BraketLocalEmulatorRoutine
    )
    with pytest.raises(ValueError):
        prog.device("foo")


def test_assign_error():
    import numpy as np

    with pytest.raises(TypeError):
        start.rydberg.detuning.uniform.constant("c", "t").assign(c=np, t=10)

    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(
            c=[1, 2, np], t=[10]
        )

    with pytest.raises(TypeError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(
            c=[1, 2, np], t=[10, 20, 30]
        )

    list_dict = [dict(c=1, t=10), dict(c=2, t=20), dict(c=np, t=30)]
    with pytest.raises(TypeError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(list_dict)

    list_dict = [dict(c=1, t=10), dict(c=2, t=20), dict(t=30)]
    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(list_dict)

    list_dict = [dict(c=1, t=10, f=1), dict(c=2, t=20, f=2), dict(t=30, c=3, f=3)]
    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(list_dict)

    dict_list = dict(
        c=[1, 2, 3], t=[10, 20, 30], mask=[[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    )

    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(**dict_list)

    list_dict = [
        dict(c=1, t=10, mask=[1, 2, 3]),
        dict(c=2, t=20, mask=[4, 5, 6]),
        dict(t=30, c=3, mask=[7, 8, 9]),
    ]

    with pytest.raises(ValueError):
        start.add_position([(0, 0), (0, 6)]).rydberg.detuning.scale("mask").constant(
            "c", "t"
        ).batch_assign(list_dict)

    # happy path is to have a list of dicts with the same keys
    start.add_position([(0, 0), (0, 6), (0, 12)]).rydberg.detuning.scale(
        "mask"
    ).constant("c", "t").batch_assign(list_dict)

    list_dict = [
        dict(c=1, t=10, mask=np.array([1, 2, 3])),
        dict(c=2, t=20, mask=np.array([4, 5, 6])),
        dict(t=30, c=3, mask=np.array([7, 8, 9])),
    ]
    with pytest.raises(ValueError):
        start.add_position([(0, 0), (0, 6)]).rydberg.detuning.scale("mask").constant(
            "c", "t"
        ).batch_assign(list_dict)

    with pytest.raises(TypeError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.piecewise_linear(
                [0.2, "rabi_dur_scanned", 0.2], [0.0, 10, 10, 0.0]
            )
            .batch_assign(rabi_dur_scanned=1.0)
        )

    with pytest.raises(TypeError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.piecewise_linear(
                [0.2, "rabi_dur_scanned", 0.2], [0.0, 10, 10, 0.0]
            )
            .assign(rabi_dur_scanned=[1.0])
        )

    with pytest.raises(TypeError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.piecewise_linear(
                [0.2, "rabi_dur_scanned", 0.2], [0.0, 10, 10, 0.0]
            )
            .amplitude.scale("rabi_mask")
            .piecewise_linear([0.2, "rabi_dur_scanned", 0.2], [0.0, 10, 10, 0.0])
            .assign(rabi_mask=0.0)
            .batch_assign(rabi_dur_scanned=[1.0])
        )


def test_flatten_dupicate_error():
    with pytest.raises(ValueError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.constant("a", 1)
            .detuning.scale("vec")
            .constant(1, 1)
            .assign(vec=[1])
            .args(["a", "a"])
            .braket.local_emulator()
            .run(1)
        )


def test_flatten_vector_error():
    with pytest.raises(ValueError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.constant(1, 1)
            .detuning.scale("a")
            .constant(1, 1)
            .args(["a"])
            .braket.local_emulator()
            .run(1)
        )


def test_apply_sequence():
    sequence = start.rydberg.detuning.uniform.constant(1, 1).parse_sequence()

    start.add_position((0, 0)).apply(sequence).braket.local_emulator().run(1)


"""
prog = start
prog = (
    prog.rydberg.detuning.location(1)
    .location(2)
    .linear(start=1.0, stop=2.0, duration="x")
    .poly(coeffs=[1, 2, 3], duration="x")
    .location(3)
    .location(4)
    .constant(value=1.0, duration="x")
    .rabi.amplitude.location(5)
    .location(6)
    .linear(start=1.0, stop=2.0, duration="x")
    .phase.location(7)
    .constant(value=1.0, duration="x")
    .apply(ir.Linear(1.0, 2.0, "x"))
    .hyperfine.detuning.location(8)
    .poly(coeffs=[1, 2, 3], duration="x")
)
print(prog)

for idx in range(5):
    prog = prog.location(idx).linear(start=1.0, stop=2.0, duration="x")

print(prog)
print(prog.sequence)


job = (
    Square(4, lattice_spacing="a")
    .apply_defect_density(0.1)
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .assign(final_detuning=20, a=4)
    .mock(100)
)
print(job)

job = (
    Square(4, lattice_spacing="a")
    .apply_defect_count(4)
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .assign(final_detuning=20, a=4)
    .mock(100)
)
print(job)


job = (
    start.add_position((0, 0))
    .add_position((6, 0))
    .add_position((3, "distance"))
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .parallelize(20.0)
    .assign(final_detuning=20, distance=4)
    .mock(100)
)
print(job)

job = (
    start.add_position((0, 0))
    .add_position((6, 0))
    .add_position((3, "distance"))
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .parallelize(20.0)
    .assign(final_detuning=20, distance=4)
    .mock(100)
)

print(job)


def my_func(time, *, omega, phi=0, amplitude):
    return amplitude * np.cos(omega * time + phi)


durations = ir.cast([0.1, "run_time", 0.1])
total_duration = sum(durations)

job = (
    start.add_position((0, 0))
    .rydberg.detuning.uniform.fn(my_func, total_duration)
    .sample(0.05, "linear")
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        durations, [0, "rabi_max", "rabi_max", 0]
    )
    .assign(omega=15, amplitude=15, rabi_max=15)
    .batch_assign(run_time=np.linspace(0, 4.0, 101))
    .braket_local_simulator(1000)
)

print(job)


n_atoms = 11
atom_spacing = 6.1
run_time = var("run_time")
run_times = np.linspace(0.1, 4.0, 11)

quantum_scar_program = (
    Chain(11, lattice_spacing=atom_spacing)
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        [0.3, 1.6, 0.3], [0.0, 15.7, 15.7, 0.0]
    )
    .piecewise_linear([0.2, 1.4, 0.2], [0, 15.7, 15.7, 0])
    .slice(stop=run_time - 0.06)
    .record("rabi_value")
    .linear("rabi_value", 0, 0.06)
    .detuning.uniform.piecewise_linear([0.3, 1.6, 0.3], [-18.8, -18.8, 16.3, 16.3])
    .piecewise_linear([0.2, 1.6], [16.3, 0.0, 0.0])
    .slice(stop=run_time)
    .batch_assign(run_time=run_times)
    .braket_local_simulator(10000)
)


sequence = (
    start.rydberg.detuning.uniform.fn(my_func, total_duration)
    .sample(0.05, "linear")
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        durations, [0, "rabi_max", "rabi_max", 0]
    )
    .sequence
)

builder = start
for site in range(11):
    builder = builder.add_position((6 * site, 0))
    job = (
        builder.apply(sequence)
        .assign(omega=15, amplitude=15, rabi_max=15)
        .batch_assign(run_time=np.linspace(0, 4.0, 101))
        .braket_local_simulator(1000)
    )
    print(job)
"""



from bloqade.ir.location import Square
import numpy as np
from decimal import Decimal
from bloqade.visualization import builder_figure
from bokeh.io import save


def test_builder_vis():
    program = (
        Square(3, lattice_spacing=1 / np.sqrt(Decimal(2)))
        .apply_defect_count(4)
        .scale(6.1)
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 0.5, 0.1], [-10, -10, "final_detuning", "final_detuning"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 0.5, 0.1], [0, 15, 15, 0])
        .batch_assign(final_detuning=np.linspace(0, 50, 51).tolist())
    )
    fig = builder_figure(program, 0)
    save(fig)



from pydantic.dataclasses import dataclass
from dataclasses import fields
import numpy as np
import itertools
from numpy.typing import NDArray
from bloqade.ir.location.base import AtomArrangement, LocationInfo
from dataclasses import InitVar
from bloqade.ir import Literal, Scalar, cast

from bloqade.builder.typing import ScalarType
from beartype.typing import List, Tuple, Generator, Optional, Any
from beartype import beartype

import plotext as pltxt
import sys


class Cell:
    @beartype
    def __init__(self, natoms: int, ndims: int) -> None:
        self.natoms = natoms
        self.ndims = ndims


@dataclass
class BoundedBravais(AtomArrangement):
    __match_args__ = ("shape", "lattice_spacing")
    """Base classe for Bravais lattices
    [`AtomArrangement`][bloqade.ir.location.base.AtomArrangement].

    - [`Square`][bloqade.ir.location.bravais.Square]
    - [`Chain`][bloqade.ir.location.bravais.Chain]
    - [`Honeycomb`][bloqade.ir.location.bravais.Honeycomb]
    - [`Triangular`][bloqade.ir.location.bravais.Triangular]
    - [`Lieb`][bloqade.ir.location.bravais.Lieb]
    - [`Kagome`][bloqade.ir.location.bravais.Kagome]
    - [`Rectangular`][bloqade.ir.location.bravais.Rectangular]


    """

    shape: Tuple[int, ...]
    lattice_spacing: Scalar

    @beartype
    def __init__(self, *shape: int, lattice_spacing: ScalarType = 1.0):
        self.shape = shape
        self.lattice_spacing = cast(lattice_spacing)
        self.__n_atoms = None
        self.__n_dims = None
        super().__init__()

    def cell_vectors(self) -> Tuple[Tuple[Scalar, ...], ...]:
        raise NotImplementedError

    def cell_atoms(self) -> Tuple[Tuple[Scalar, ...], ...]:
        raise NotImplementedError

    def __str__(self):
        has_lattice_spacing_var = False
        if type(self.lattice_spacing) is not Literal:
            # add string denoting this to printer
            repr_lattice_spacing = 1.0
            has_lattice_spacing_var = True
        else:
            repr_lattice_spacing = float(self.lattice_spacing.value)

        xs, ys = [], []

        for index in itertools.product(*[range(n) for n in self.shape]):
            for pos in self.coordinates(list(index)):
                (x, y) = tuple(repr_lattice_spacing * pos)
                xs.append(x)
                ys.append(y)

        pltxt.clear_figure()
        pltxt.limit_size(False, False)
        pltxt.plot_size(80, 24)
        pltxt.canvas_color("default")
        pltxt.axes_color("default")
        pltxt.ticks_color("white")
        pltxt.title("Atom Positions")
        pltxt.xlabel("x (um)")
        pltxt.ylabel("y (um)")
        if has_lattice_spacing_var:
            pltxt.ylabel(
                "Lattice Spacing is a variable, defaulting to 1.0 for display", "right"
            )

        marker = "."
        if sys.stdout.encoding.lower().startswith("utf"):
            marker = "dot"

        pltxt.scatter(xs, ys, color=(100, 55, 255), marker=marker)

        return pltxt.build()

    @property
    def n_atoms(self) -> int:
        if not self.__n_atoms:
            self.__n_atoms = len(self.cell_atoms()) * np.prod(self.shape)
        return self.__n_atoms

    @property
    def n_sites(self) -> int:
        return self.n_atoms

    @property
    def n_vacant(self) -> int:
        return 0

    @property
    def n_dims(self):
        """dimension of the lattice

        Returns:
            int: dimension of the lattice

        """
        if not self.__n_dims:
            self.__n_dims = len(self.cell_vectors())
        return self.__n_dims

    @beartype
    def coordinates(self, index: List[int]) -> NDArray:
        """calculate the coordinates of a cell in the lattice
        given the cell index.
        """
        # damn! this is like stone age broadcasting
        vectors = np.array(self.cell_vectors())
        index = np.array(index)
        pos = np.sum(vectors.T * index, axis=1)
        return pos + np.array(self.cell_atoms())

    def enumerate(self) -> Generator[LocationInfo, None, None]:
        for index in itertools.product(*[range(n) for n in self.shape]):
            for pos in self.coordinates(list(index)):
                position = tuple(self.lattice_spacing * pos)
                yield LocationInfo(position, True)

    def __iter__(self):
        for index in itertools.product(*[range(n) for n in self.shape]):
            for pos in self.coordinates(list(index)):
                position = tuple(self.lattice_spacing * pos)
                yield LocationInfo(position, True)

    @beartype
    def scale(self, factor: ScalarType) -> "BoundedBravais":
        """Scale the current location with a factor.

        (x,y) -> factor*(x,y)

        Args:
            factor (str | Real | Decimal | Scalar): scale factor

        Returns:
            BoundedBravais: The lattice with the scaled locations
        """
        factor = cast(factor)
        obj = self.__new__(type(self))
        for f in fields(self):
            if f.name == "lattice_spacing":
                obj.lattice_spacing = factor * self.lattice_spacing
            else:
                setattr(obj, f.name, getattr(self, f.name))
        return obj


@dataclass(init=False)
class Chain(BoundedBravais):
    """Chain lattice.

    - 1D lattice
    - primitive (cell) vector(s)
        - a1 = (1,0).
    - unit cell (1 atom(s))
        - loc (0,0)

    Args:
        L (int): number of sites in the chain
        lattice_spacing (Scalar, Real): lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    vertical: bool
    vertical_chain: InitVar[bool]

    @beartype
    def __init__(
        self, L: int, lattice_spacing: ScalarType = 1.0, vertical_chain: bool = False
    ):
        self.vertical = vertical_chain
        super().__init__(L, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        if self.vertical:
            return [[0, 1]]
        else:
            return [[1, 0]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0, 0]]


@dataclass
class Square(BoundedBravais):
    """Square lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1,0)
        - a2 = (0,1)
    - unit cell (1 atom(s))
        - loc (0,0)

    Args:
        L1 (int): number of sites in linear direction. n_atoms = L1 * L1.
        L2 (Optional[int]): number of sites in direction a2.
            n_atoms = L1 * L2, default is L1
        lattice_spacing (Scalar, Real): lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    @beartype
    def __init__(
        self, L1: int, L2: Optional[int] = None, lattice_spacing: ScalarType = 1.0
    ):
        if L2 is None:
            L2 = L1
        super().__init__(L1, L2, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        return [[1, 0], [0, 1]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0, 0]]


@dataclass(init=False)
class Rectangular(BoundedBravais):
    """Rectangular lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1,0)
        - a2 = (0,1)
    - unit cell (1 atom(s))
        - loc (0,0)


    Args:
        width (int): number of sites in x direction.
        height (int): number of sites in y direction.
        lattice_spacing_x (Scalar, Real):
            lattice spacing. Defaults to 1.0.
        lattice_spacing_y (Scalar, Real):
            lattice spacing in y direction. optional.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    __match_args__ = ("shape", "lattice_spacing", "ratio")
    ratio: Scalar = 1.0
    lattice_spacing_x: InitVar[Any]
    lattice_spacing_y: InitVar[Any]

    @beartype
    def __init__(
        self,
        width: int,
        height: int,
        lattice_spacing_x: ScalarType = 1.0,
        lattice_spacing_y: Optional[ScalarType] = None,
    ):
        super().__init__(width, height, lattice_spacing=lattice_spacing_x)

        if lattice_spacing_y is None:
            self.ratio = cast(1.0) / cast(lattice_spacing_x)
        else:
            self.ratio = cast(lattice_spacing_y) / cast(lattice_spacing_x)

        super().__init__(width, height, lattice_spacing=lattice_spacing_x)

    def __str__(self):
        # modified version of the standard coordinates method,
        # intercept cell.vectors, then continue with standard
        # operation
        def repr_compatible_coordinates(self, index: List[int]) -> NDArray:
            cell_vectors = self.cell_vectors()
            cell_vectors[1][1] = float(cell_vectors[1][1].value)

            vectors = np.array(cell_vectors)
            index = np.array(index)
            pos = np.sum(vectors.T * index, axis=1)
            return pos + np.array(self.cell_atoms())

        # if ratio is a Literal, then
        # by extension the lattice_spacing must also be a Literal
        if isinstance(self.ratio, Literal):
            repr_lattice_spacing = float(self.lattice_spacing.value)

            xs, ys = [], []

            for index in itertools.product(*[range(n) for n in self.shape]):
                for pos in repr_compatible_coordinates(
                    self, index
                ):  # need to replace this
                    (x, y) = tuple(repr_lattice_spacing * pos)
                    xs.append(x)
                    ys.append(y)

            pltxt.clear_figure()
            pltxt.limit_size(False, False)
            pltxt.plot_size(80, 24)
            pltxt.canvas_color("default")
            pltxt.axes_color("default")
            pltxt.ticks_color("white")
            pltxt.title("Atom Positions")
            pltxt.xlabel("x (um)")
            pltxt.ylabel("y (um)")

            marker = "."
            if sys.stdout.encoding.lower().startswith("utf"):
                marker = "dot"

            pltxt.scatter(xs, ys, color=(100, 55, 255), marker=marker)

            return pltxt.build()

        else:
            return "Rectangular(shape={}, lattice_spacing={}, ratio={})".format(
                self.shape, self.lattice_spacing, self.ratio
            )

    def cell_vectors(self) -> List[List[float]]:
        return [[1, 0], [0, self.ratio]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0, 0]]


@dataclass
class Honeycomb(BoundedBravais):
    """Honeycomb lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1, 0)
        - a2 = (1/2, sqrt(3)/2)
    - unit cell (2 atom(s))
        - loc1 (0, 0)
        - loc2 (1/2, 1/(2*sqrt(3))


    Args:
        L1 (int): number of unit cells in linear direction. n_atoms = L1 * L1 * 2.
        L2 (Optional[int]): number of unit cells in direction a2.
            n_atoms = L1 * L2 * 2, default is L1.

        lattice_spacing (Scalar, Real):
            lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    @beartype
    def __init__(
        self, L1: int, L2: Optional[int] = None, lattice_spacing: ScalarType = 1.0
    ):
        if L2 is None:
            L2 = L1
        super().__init__(L1, L2, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        return [[1.0, 0.0], [1 / 2, np.sqrt(3) / 2]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0.0, 0.0], [1 / 2, 1 / (2 * np.sqrt(3))]]


@dataclass
class Triangular(BoundedBravais):
    """Triangular lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1, 0)
        - a2 = (1/2, sqrt(3)/2)
    - unit cell (1 atom(s))
        - loc (0, 0)


    Args:
        L (int): number of sites in linear direction. n_atoms = L * L.
        L2 (Optional[int]): number of sites along a2 direction,
            n_atoms = L1 * L2, default is L1.
        lattice_spacing (Scalar, Real):
            lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    @beartype
    def __init__(
        self, L1: int, L2: Optional[int] = None, lattice_spacing: ScalarType = 1.0
    ):
        if L2 is None:
            L2 = L1
        super().__init__(L1, L2, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        return [[1.0, 0.0], [1 / 2, np.sqrt(3) / 2]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0.0, 0.0]]


@dataclass
class Lieb(BoundedBravais):
    """Lieb lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1, 0)
        - a2 = (0, 1)
    - unit cell (3 atom(s))
        - loc1 (0, 0)
        - loc2 (0.5, 0)
        - loc3 (0 ,0.5)

    Args:
        L1 (int): number of unit cells in linear direction. n_atoms = 3* L1 * L1.
        L2 (Optional[int]): number of unit cells along a2 direction,
            n_atoms = 3 * L1 * L2, default is L1.
        lattice_spacing (Scalar, Real):
            lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    @beartype
    def __init__(
        self, L1: int, L2: Optional[int] = None, lattice_spacing: ScalarType = 1.0
    ):
        if L2 is None:
            L2 = L1
        super().__init__(L1, L2, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        return [[1.0, 0.0], [0.0, 1.0]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0.0, 0.0], [1 / 2, 0.0], [0.0, 1 / 2]]


@dataclass
class Kagome(BoundedBravais):
    """Kagome lattice.

    - 2D lattice
    - primitive (cell) vector(s)
        - a1 = (1, 0)
        - a2 = (1/2, sqrt(3)/2)
    - unit cell (3 atom(s))
        - loc1 (0, 0)
        - loc2 (0.5, 0)
        - loc3 (0.25 ,0.25sqrt(3))

    Args:
        L1 (int): number of sites in linear direction. n_atoms = 3 * L1 * L1.
        L2 (Optional[int]): number of unit cells along a2 direction,
            n_atoms = 3 * L1 * L2, default is L1.
        lattice_spacing (Scalar, Real):
            lattice spacing. Defaults to 1.0.


    - Possible Next:
        continue with `.` to see possible next step in auto-prompt
        supported setting (IPython, IDE ...)

    """

    @beartype
    def __init__(
        self, L1: int, L2: Optional[int] = None, lattice_spacing: ScalarType = 1.0
    ):
        if L2 is None:
            L2 = L1
        super().__init__(L1, L2, lattice_spacing=lattice_spacing)

    def __repr__(self):
        return super().__repr__()

    def cell_vectors(self) -> List[List[float]]:
        return [[1.0, 0.0], [1 / 2, np.sqrt(3) / 2]]

    def cell_atoms(self) -> List[List[float]]:
        return [[0.0, 0.0], [1 / 2, 0], [1 / 4, np.sqrt(3) / 4]]



from bloqade.ir.scalar import (
    Scalar,
    Literal,
    Variable,
    AssignedVariable,
    Negative,
    Add,
    Mul,
    Div,
    Max,
    Min,
    Slice,
    Interval,
)
from typing import Any


class ScalarVisitor:
    def visit_literal(self, ast: Literal) -> Any:
        raise NotImplementedError(
            f"visit_literal not implemented for {self.__class__.__name__}"
        )

    def visit_variable(self, ast: Variable) -> Any:
        raise NotImplementedError(
            f"visit_variable not implemented for {self.__class__.__name__}"
        )

    def visit_assigned_variable(self, ast: AssignedVariable) -> Any:
        raise NotImplementedError(
            f"visit_default_variable not implemented for {self.__class__.__name__}"
        )

    def visit_negative(self, ast: Negative) -> Any:
        raise NotImplementedError(
            f"visit_negative not implemented for {self.__class__.__name__}"
        )

    def visit_add(self, ast: Add) -> Any:
        raise NotImplementedError(
            f"visit_add not implemented for {self.__class__.__name__}"
        )

    def visit_mul(self, ast: Mul) -> Any:
        raise NotImplementedError(
            f"visit_mul not implemented for {self.__class__.__name__}"
        )

    def visit_div(self, ast: Div) -> Any:
        raise NotImplementedError(
            f"visit_div not implemented for {self.__class__.__name__}"
        )

    def visit_max(self, ast: Max) -> Any:
        raise NotImplementedError(
            f"visit_max not implemented for {self.__class__.__name__}"
        )

    def visit_min(self, ast: Min) -> Any:
        raise NotImplementedError(
            f"visit_min not implemented for {self.__class__.__name__}"
        )

    def visit_slice(self, ast: Slice) -> Any:
        raise NotImplementedError(
            f"visit_slice not implemented for {self.__class__.__name__}"
        )

    def visit_interval(self, ast: Interval) -> Any:
        raise NotImplementedError(
            f"visit_interval not implemented for {self.__class__.__name__}"
        )

    def visit(self, ast: Scalar) -> Any:
        if isinstance(ast, Literal):
            return self.visit_literal(ast)
        elif isinstance(ast, Variable):
            return self.visit_variable(ast)
        elif isinstance(ast, AssignedVariable):
            return self.visit_assigned_variable(ast)
        elif isinstance(ast, Negative):
            return self.visit_negative(ast)
        elif isinstance(ast, Add):
            return self.visit_add(ast)
        elif isinstance(ast, Mul):
            return self.visit_mul(ast)
        elif isinstance(ast, Div):
            return self.visit_div(ast)
        elif isinstance(ast, Max):
            return self.visit_max(ast)
        elif isinstance(ast, Min):
            return self.visit_min(ast)
        elif isinstance(ast, Slice):
            return self.visit_slice(ast)
        elif isinstance(ast, Interval):
            return self.visit_interval(ast)
        else:
            raise NotImplementedError(f"unknown ast type {ast.__class__.__name__}")



import pytest
from bloqade.ir.location import Square
from bloqade import start


# create lattice
def test_parallel_task():
    lattice = Square(3, lattice_spacing=6)

    quantum_batch = (
        lattice.rydberg.detuning.uniform.piecewise_linear(
            durations=["up_time", "anneal_time", "up_time"],
            values=[
                "initial_detuning",
                "initial_detuning",
                "final_detuning",
                "final_detuning",
            ],
        )
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=["up_time", "anneal_time", "up_time"],
            values=[0, "rabi_amplitude_max", "rabi_amplitude_max", 0],
        )
        .assign(
            initial_detuning=-10,
            up_time=0.1,
            final_detuning=15,
            anneal_time=10,
            rabi_amplitude_max=15,
        )
        .parallelize(10.0)
        .quera.mock()
        ._compile(shots=10)
    )

    assert quantum_batch.tasks[0].parallel_decoder


def test_error_parallel_noatom():
    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.piecewise_linear(
            durations=["up_time", "anneal_time", "up_time"],
            values=[
                "initial_detuning",
                "initial_detuning",
                "final_detuning",
                "final_detuning",
            ],
        ).parallelize(10.0).quera.mock()._compile(shots=1)



from bloqade.codegen.common.json import WaveformSerializer, BloqadeIRDeserializer
from bloqade.serialize import Serializer
from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Dict, List, Tuple, Optional
from enum import Enum
from bloqade.ir.control.waveform import Waveform
from bloqade.emulate.ir.atom_type import AtomType


@dataclass(frozen=True)
@Serializer.register
class CompiledWaveform:
    assignments: Dict[str, Decimal]
    source: Waveform

    def __call__(self, t: float) -> float:
        return self.source(t, **self.assignments)


@CompiledWaveform.set_serializer
def _serialize(obj: CompiledWaveform) -> Dict[str, Any]:
    return {
        "assignments": obj.assignments,
        "source": WaveformSerializer().default(obj.source),
    }


@CompiledWaveform.set_deserializer
def _deserializer(d: Dict[str, Any]) -> CompiledWaveform:
    from json import loads, dumps

    source_str = dumps(d["source"])
    d["source"] = loads(source_str, object_hook=BloqadeIRDeserializer.object_hook)
    return CompiledWaveform(**d)


class RabiOperatorType(int, Enum):
    RabiAsymmetric = 0
    RabiSymmetric = 1


@dataclass(frozen=True)
@Serializer.register
class RabiOperatorData:
    operator_type: RabiOperatorType
    target_atoms: Dict[int, Decimal]

    def __hash__(self):
        return hash(self.operator_type) ^ hash(frozenset(self.target_atoms.items()))


@dataclass(frozen=True)
@Serializer.register
class RabiTerm:
    operator_data: RabiOperatorData
    amplitude: CompiledWaveform
    phase: Optional[CompiledWaveform] = None


@dataclass(frozen=True)
@Serializer.register
class DetuningOperatorData:
    target_atoms: Dict[int, Decimal]

    def __hash__(self) -> int:
        return hash(frozenset(self.target_atoms.items()))


@dataclass(frozen=True)
@Serializer.register
class DetuningTerm:
    operator_data: DetuningOperatorData
    amplitude: CompiledWaveform


@dataclass(frozen=True)
@Serializer.register
class Fields:
    detuning: List[DetuningTerm]
    rabi: List[RabiTerm]


@dataclass(frozen=True)
@Serializer.register
class Register:
    """This class represents the of the atoms in the system."""

    atom_type: AtomType
    sites: List[Tuple[Decimal, Decimal]]
    blockade_radius: Decimal

    def __len__(self):
        return len(self.sites)

    def __eq__(self, other: Any):
        if isinstance(other, Register):
            return (
                self.atom_type == other.atom_type
                and self.blockade_radius == other.blockade_radius
                and set(self.sites) == set(other.sites)
            )

        return False

    def __hash__(self) -> int:
        if self.blockade_radius == Decimal("0"):
            # if blockade radius is zero, then the positions are irrelevant
            # because the fock states generated by the geometry are the same
            return hash(self.__class__) ^ hash(self.atom_type) ^ hash(len(self.sites))
        else:
            # if blockade radius is non-zero, then the positions are relevant
            # because depending on the location of the atoms and the blockade
            # radius, the fock states generated by the geometry are different
            return (
                hash(self.__class__)
                ^ hash(self.atom_type)
                ^ hash(self.blockade_radius)
                ^ hash(frozenset(self.sites))
            )


class LevelCoupling(str, Enum):
    Rydberg = "rydberg"
    Hyperfine = "hyperfine"


@dataclass(frozen=True)
@Serializer.register
class EmulatorProgram:
    register: Register
    duration: float
    pulses: Dict[LevelCoupling, Fields]


class Visitor:
    def visit_emulator_program(self, ast: EmulatorProgram) -> Any:
        raise NotImplementedError

    def visit_compiled_waveform(self, ast: CompiledWaveform) -> Any:
        raise NotImplementedError

    def visit_fields(self, ast: Fields) -> Any:
        raise NotImplementedError

    def visit_detuning_operator_data(self, ast: DetuningOperatorData) -> Any:
        raise NotImplementedError

    def visit_rabi_operator_data(self, ast: RabiOperatorData) -> Any:
        raise NotImplementedError

    def visit_detuning_term(self, ast: DetuningTerm) -> Any:
        raise NotImplementedError

    def visit_rabi_term(self, ast: RabiTerm) -> Any:
        raise NotImplementedError

    def visit_register(self, ast: Register) -> Any:
        raise NotImplementedError

    def visit(self, ast) -> Any:
        if isinstance(ast, EmulatorProgram):
            return self.visit_emulator_program(ast)
        elif isinstance(ast, Register):
            return self.visit_register(ast)
        elif isinstance(ast, Fields):
            return self.visit_fields(ast)
        elif isinstance(ast, RabiTerm):
            return self.visit_rabi_term(ast)
        elif isinstance(ast, DetuningTerm):
            return self.visit_detuning_term(ast)
        elif isinstance(ast, RabiOperatorData):
            return self.visit_rabi_operator_data(ast)
        elif isinstance(ast, DetuningOperatorData):
            return self.visit_detuning_operator_data(ast)
        elif isinstance(ast, CompiledWaveform):
            return self.visit_compiled_waveform(ast)
        else:
            raise NotImplementedError(f"Unknown AST node type {type(ast)}: {ast!r}")
